#!/usr/bin/env python3

import itertools
import logging
import unittest

from somajo import Tokenizer
from somajo.doubly_linked_list import DLL
from somajo.token import Token
from somajo import utils


class TestTokenizer(unittest.TestCase):
    def setUp(self):
        """Necessary preparations"""
        self.tokenizer = Tokenizer(language="de_CMC", split_camel_case=True)

    def _equal(self, raw, tokenized):
        """"""
        if isinstance(tokenized, str):
            tokenized = tokenized.split()
        dll = DLL([Token(raw, first_in_sentence=True, last_in_sentence=True)])
        tokens = self.tokenizer._tokenize(dll)
        self.assertEqual([t.text for t in tokens], tokenized)

    def _equal_xml(self, raw, tokenized):
        """"""
        if isinstance(tokenized, str):
            tokenized = tokenized.split()
        eos_tags = "title h1 h2 h3 h4 h5 h6 p br hr div ol ul dl table".split()
        eos_tags = set(eos_tags)
        chunk_info = utils.xml_chunk_generator(raw, is_file=False, eos_tags=eos_tags)
        chunk_lists = (ci[0] for ci in chunk_info)
        token_dlls = map(DLL, chunk_lists)
        chunks = map(self.tokenizer._tokenize, token_dlls)
        complete = list(itertools.chain.from_iterable(chunks))
        complete = utils.escape_xml_tokens(complete)
        self.assertEqual([t.text for t in complete], tokenized)


class TestEnglishTokenizer(TestTokenizer):
    def setUp(self):
        """Necessary preparations"""
        self.tokenizer = Tokenizer(language="en_PTB", split_camel_case=True)


class TestTokenizerDeprecated(TestTokenizer):
    def _equal(self, raw, tokenized):
        if isinstance(tokenized, str):
            tokenized = tokenized.split()
        with self.assertLogs(level=logging.WARN):
            tokens = self.tokenizer.tokenize_paragraph(raw)
        self.assertEqual(tokens, tokenized)

    def _equal_xml(self, raw, tokenized):
        if isinstance(tokenized, str):
            tokenized = tokenized.split()
        with self.assertLogs(level=logging.WARN):
            tokens = self.tokenizer.tokenize_xml(raw, is_file=False)
        self.assertEqual(tokens, tokenized)


class TestGuidelines(TestTokenizer):
    """Test cases taken from the EmpiriST tokenization guidelines:
    https://docs.google.com/viewer?a=v&pid=sites&srcid=ZGVmYXVsdGRvbWFpbnxlbXBpcmlzdDIwMTV8Z3g6M2Y3MDlkYmQzMDk4NzFlZA

    Function names: test_<category>_<page><n>

    """

    def test_punctuation_0201(self):
        self._equal("Als ich ihn sah, war es bereits zu spät.", "Als ich ihn sah , war es bereits zu spät .")

    def test_punctuation_0601(self):
        self._equal("Das hab ich mich auch schon gefragt...", "Das hab ich mich auch schon gefragt ...")

    def test_punctuation_0602(self):
        self._equal("Das ist ein Test?!", "Das ist ein Test ?!")

    def test_punctuation_0603(self):
        self._equal("bh's", "bh's")

    def test_punctuation_0701(self):
        self._equal("mit'm Fahrrad", "mit'm Fahrrad")

    def test_punctuation_0702(self):
        self._equal("des Virus'", "des Virus'")

    @unittest.expectedFailure
    def test_punctuation_0703(self):
        self._equal("du bist echt ein a...", "du bist echt ein a...")

    @unittest.expectedFailure
    def test_punctuation_0704(self):
        self._equal("f*** you!", "f*** you !")

    def test_punctuation_0705(self):
        self._equal("Test.......", "Test .......")

    def test_punctuation_0706(self):
        self._equal("Test????????", "Test ????????")

    def test_punctuation_0707(self):
        self._equal("Test!!!", "Test !!!")

    def test_punctuation_0708(self):
        self._equal("Test?!?!?!", "Test ?!?!?!")

    def test_punctuation_0801(self):
        self._equal("In der 2....", "In der 2. ...")

    def test_punctuation_0802(self):
        self._equal("Test....????!!!!", "Test .... ????!!!!")

    def test_punctuation_0803(self):
        self._equal("Laub- und Nadelbäume", "Laub- und Nadelbäume")

    def test_punctuation_0804(self):
        self._equal("Hals-Nasen-Ohren-Arzt", "Hals-Nasen-Ohren-Arzt")

    def test_punctuation_0901(self):
        self._equal("10%", "10 %")

    def test_punctuation_0902(self):
        self._equal("200€", "200 €")

    def test_punctuation_0903(self):
        self._equal("§§48", "§§ 48")

    def test_punctuation_0904(self):
        self._equal("11+21=33", "11 + 21 = 33")

    def test_punctuation_0905(self):
        self._equal("Avril_Lavigne", "Avril_Lavigne")

    @unittest.expectedFailure
    def test_punctuation_0906(self):
        self._equal("+s", "+s")

    def test_punctuation_0907(self):
        self._equal("-v", "-v")

    def test_punctuation_1001(self):
        self._equal("f->d", "f -> d")

    def test_punctuation_1001a(self):
        self._equal("f→d", "f → d")

    def test_punctuation_1002(self):
        self._equal("f - > d", "f -> d")

    @unittest.expectedFailure
    def test_punctuation_1003(self):
        self._equal("/ig schdöbbs", "/ig schdöbbs")

    def test_punctuation_1004(self):
        self._equal("(Neu-)Veröffentlichung", "( Neu- ) Veröffentlichung")

    def test_punctuation_1005(self):
        self._equal("Student(inn)en", "Student(inn)en")

    def test_punctuation_1101(self):
        self._equal("StudentInnen", "StudentInnen")

    def test_punctuation_1102(self):
        self._equal("Student/innen", "Student/innen")

    def test_punctuation_1103(self):
        self._equal("Experten/Expertinnen", "Experten / Expertinnen")

    def test_punctuation_1104(self):
        self._equal("(“normale”/saisonale) Grippe", "( “ normale ” / saisonale ) Grippe")

    def test_punctuation_1105(self):
        self._equal("Vitamin C-Mangel", "Vitamin C-Mangel")

    def test_punctuation_1106(self):
        self._equal("Magen- Darm- Erkrankung", "Magen- Darm- Erkrankung")

    def test_punctuation_1201(self):
        self._equal("Oster-„Beckerei“", "Oster- „ Beckerei “")

    def test_punctuation_1202(self):
        self._equal("i-„Pott“", "i- „ Pott “")

    def test_sumtimedate_1301(self):
        self._equal("6.200", "6.200")

    def test_sumtimedate_1302(self):
        self._equal("6.200,-", "6.200,-")

    def test_sumtimedate_1303(self):
        self._equal("das dauert 2:40 std.", "das dauert 2:40 std.")

    def test_sumtimedate_1304(self):
        self._equal("-1,5", "-1,5")

    def test_sumtimedate_1304a(self):
        """Unicode minus sign"""
        self._equal("−1,5", "−1,5")

    def test_sumtimedate_1305(self):
        self._equal("1/2 h", "1/2 h")

    def test_sumtimedate_1306(self):
        self._equal("1 / 2 h", "1 / 2 h")

    def test_sumtimedate_1307(self):
        self._equal("14:00–18:00", "14:00 – 18:00")

    def test_sumtimedate_1401(self):
        self._equal("14–18 Uhr", "14 – 18 Uhr")

    def test_sumtimedate_1402(self):
        self._equal("1-3 Semester", "1 - 3 Semester")

    def test_sumtimedate_1403(self):
        self._equal("30-50kg", "30 - 50 kg")

    def test_sumtimedate_1404(self):
        self._equal("ca. 20°C", "ca. 20 ° C")

    def test_sumtimedate_1404a(self):
        self._equal("ca. 20 °C", "ca. 20 ° C")

    def test_sumtimedate_1405(self):
        self._equal("ca. 20 GHz", "ca. 20 GHz")

    @unittest.expectedFailure
    def test_sumtimedate_1406(self):
        self._equal("4-11mal", "4 - 11mal")

    def test_sumtimedate_1501(self):
        self._equal("16.07.2013", "16. 07. 2013")

    def test_sumtimedate_1502(self):
        self._equal("16. Juli 2013", "16. Juli 2013")

    def test_sumtimedate_1502a(self):
        self._equal("(am 20.06.2008)", "( am 20. 06. 2008 )")

    def test_sumtimedate_1503(self):
        self._equal("21/07/1980", "21/ 07/ 1980")

    def test_sumtimedate_1503a(self):
        self._equal("2016-01-27", "2016 -01 -27")

    def test_sumtimedate_1503b(self):
        self._equal("01-27-2016", "01- 27- 2016")

    def test_sumtimedate_1504(self):
        self._equal("Weimarer Klassik", "Weimarer Klassik")

    def test_sumtimedate_1505(self):
        self._equal("Hälfte des XIX Jh. = Anfang 1796", "Hälfte des XIX Jh. = Anfang 1796")

    def test_sumtimedate_1601(self):
        self._equal("WS04", "WS 04")

    def test_sumtimedate_1601a(self):
        self._equal("WS15/16", "WS 15/16")

    def test_sumtimedate_1602(self):
        self._equal("24-stündig", "24-stündig")

    def test_abbreviations_1603(self):
        self._equal("etc.", "etc.")

    def test_abbreviations_1604(self):
        self._equal("d.h.", "d. h.")

    def test_abbreviations_1605(self):
        self._equal("d. h.", "d. h.")

    def test_abbreviations_1606(self):
        self._equal("o.ä.", "o. ä.")

    def test_abbreviations_1607(self):
        self._equal("u.dgl.", "u. dgl.")

    def test_abbreviations_1608(self):
        self._equal("u. dgl.", "u. dgl.")

    def test_abbreviations_1609(self):
        self._equal("std.", "std.")

    def test_abbreviations_1610(self):
        self._equal("Mat.-Nr.", "Mat.-Nr.")

    def test_abbreviations_1611(self):
        self._equal("Wir kauften Socken, Hemden, Schuhe etc. Danach hatten wir alles.", "Wir kauften Socken , Hemden , Schuhe etc. Danach hatten wir alles .")

    def test_abbreviations_1701(self):
        self._equal("zB", "zB")

    def test_abbreviations_1702(self):
        self._equal("Zeitschr.titel", "Zeitschr.titel")

    def test_abbreviations_1703(self):
        self._equal("A9", "A9")

    def test_abbreviations_1704(self):
        self._equal("cu Peter", "cu Peter")

    def test_abbreviations_1705(self):
        self._equal("Bruce Springsteen aka The Boss", "Bruce Springsteen aka The Boss")

    def test_typos_1801(self):
        self._equal("die stehen da schona ber ohne preise", "die stehen da schona ber ohne preise")

    def test_typos_1802(self):
        self._equal("tag quaki : )", "tag quaki :)")

    def test_typos_1901(self):
        self._equal("Ich hab da noch maldrüber nachgedacht", "Ich hab da noch maldrüber nachgedacht")

    def test_typos_1902(self):
        self._equal("Anna,kannst du mal", "Anna , kannst du mal")

    @unittest.expectedFailure
    def test_typos_1903(self):
        self._equal("handfest un direkt- so sind se...die Pottler", "handfest un direkt - so sind se ... die Pottler")

    def test_typos_1904(self):
        self._equal("Warst du vom Zeugnis überraschtß", "Warst du vom Zeugnis überraschtß")

    def test_contractions_2001(self):
        self._equal("stimmt’s", "stimmt’s")

    def test_contractions_2002(self):
        self._equal("waren’s", "waren’s")

    def test_camel_case_2101(self):
        self._equal("Zu welchemHandlungsbereich gehört unsereKomm hier? Bildung?Freizeit?Mischung?", "Zu welchem Handlungsbereich gehört unsere Komm hier ? Bildung ? Freizeit ? Mischung ?")

    def test_camel_case_2102(self):
        self._equal("derText", "der Text")

    def test_camel_case_2103(self):
        self._equal("PepsiCo", "PepsiCo")

    def test_tags_2104(self):
        self._equal('eine <A target="_blank" href="https://en.wikipedia.org/w/index.php?tit-le=Talk:PRISM_(surveillance_program)&oldid=559238329#Known_Counter_Measures_deleted_.21">umfangreiche Diskussion</A> zu diesem Abschnitt', ['eine', '<A target="_blank" href="https://en.wikipedia.org/w/index.php?tit-le=Talk:PRISM_(surveillance_program)&oldid=559238329#Known_Counter_Measures_deleted_.21">', 'umfangreiche', 'Diskussion', '</A>', 'zu', 'diesem', 'Abschnitt'])

    def test_emails_urls_2201(self):
        self._equal("michael.beisswenger@tu-dortmund.de", "michael.beisswenger@tu-dortmund.de")

    def test_emails_urls_2202(self):
        self._equal("https://en.wikipedia.org/wiki/Main_Page", "https://en.wikipedia.org/wiki/Main_Page")

    def test_emails_urls_2203(self):
        self._equal("http://www.shortnews.de", "http://www.shortnews.de")

    def test_emails_urls_2203a(self):
        self._equal("Auf http://www.shortnews.de, der besten Seite.", "Auf http://www.shortnews.de , der besten Seite .")

    def test_emails_urls_2203b(self):
        self._equal("Auf www.shortnews.de, der besten Seite.", "Auf www.shortnews.de , der besten Seite .")

    def test_emails_urls_2204(self):
        self._equal("shortnews.de", "shortnews.de")

    def test_emails_urls_2204a(self):
        self._equal("Auf shortnews.de, der besten Seite.", "Auf shortnews.de , der besten Seite .")

    def test_emoticons_2205(self):
        self._equal(":-)", ":-)")

    def test_emoticons_2206(self):
        self._equal(":-)))))", ":-)))))")

    def test_emoticons_2207(self):
        self._equal(";-)", ";-)")

    def test_emoticons_2208(self):
        self._equal(":)", ":)")

    def test_emoticons_2209(self):
        self._equal(";)", ";)")

    def test_emoticons_2301(self):
        self._equal(":-(", ":-(")

    def test_emoticons_2302(self):
        self._equal("8)", "8)")

    def test_emoticons_2303(self):
        self._equal(":D", ":D")

    def test_emoticons_2304(self):
        self._equal("^^", "^^")

    def test_emoticons_2305(self):
        self._equal("o.O", "o.O")

    def test_emoticons_2306(self):
        self._equal("oO", "oO")

    def test_emoticons_2307(self):
        self._equal("\\O/", "\\O/")

    def test_emoticons_2308(self):
        self._equal("\\m/", "\\m/")

    def test_emoticons_2309(self):
        self._equal(":;))", ":;))")

    def test_emoticons_2310(self):
        self._equal("_))", "_))")

    def test_emoticons_2311(self):
        self._equal("tag quaki : )", "tag quaki :)")

    def test_emoticons_2311a(self):
        self._equal("hallo peter ; )", "hallo peter ;)")

    def test_emoticons_2312(self):
        self._equal("Das hab ich auch schon gehört (find ich super :-)", "Das hab ich auch schon gehört ( find ich super :-)")

    def test_actions_2401(self):
        self._equal("*grübel*", "* grübel *")

    def test_actions_2402(self):
        self._equal("*auf locher rumhüpf & konfetti mach*", "* auf locher rumhüpf & konfetti mach *")

    def test_actions_2501(self):
        self._equal("*dichmalganzdolleknuddelt*", "* dichmalganzdolleknuddelt *")

    def test_actions_2502(self):
        self._equal("immer noch nicht fassen kann", "immer noch nicht fassen kann")

    def test_actions_2503(self):
        self._equal("quakirenntdemflöppymitdeneiswürfelnnach*", "quakirenntdemflöppymitdeneiswürfelnnach *")

    def test_actions_2504(self):
        self._equal("+s*", "+ s *")

    def test_actions_2505(self):
        self._equal("*danachdenraminmeinenrechnereinbau*G*", "* danachdenraminmeinenrechnereinbau * G *")

    def test_actions_2601(self):
        self._equal("quaki knuddelt Thor", "quaki knuddelt Thor")

    def test_addressing_2602(self):
        self._equal("@bine23", "@bine23")

    def test_addressing_2603(self):
        self._equal("@alle", "@alle")

    def test_addressing_2604(self):
        self._equal("winke@bochum", "winke @bochum")

    def test_addressing_2605(self):
        self._equal("@bine23: hallöchen! :-)", "@bine23 : hallöchen ! :-)")

    def test_hashtags_2701(self):
        self._equal("#urlaub", "#urlaub")

    def test_hashtags_2702(self):
        self._equal("#SPD", "#SPD")


class TestWhitespace(TestTokenizer):
    def test_whitespace_01(self):
        # self.assertEqual(self.tokenizer.tokenize_text("Petra und Simone gehen ins Kino"), "Petra und Simone gehen ins Kino".split())
        self._equal("Petra und Simone gehen ins Kino", "Petra und Simone gehen ins Kino")

    def test_whitespace_02(self):
        # newline
        self._equal("foo\nbar", "foo bar")

    def test_whitespace_03(self):
        # Unicode line separator (2028)
        self._equal("foo\u2028bar", "foo bar")

    def test_whitespace_04(self):
        # Unicode paragraph separator (2029)
        self._equal("foo\u2029bar", "foo bar")


class TestJunk(TestTokenizer):
    def test_junk_01(self):
        # zero width space
        self._equal("foo​bar", "foobar")

    def test_junk_02(self):
        # soft hyphen
        self._equal("foo­bar", "foobar")

    def test_junk_03(self):
        # zero-width no-break space (FEFF)
        self._equal("foo\ufeffbar", "foobar")

    def test_junk_04(self):
        # control characters
        self._equal("foobarbazquxalphabetagamma", "foobarbazquxalphabetagamma")

    def test_junk_05(self):
        # zero width joiner and non-joiner
        self._equal("foo‌bar‍baz", "foobarbaz")

    def test_junk_06(self):
        # left-to-right and right-to-left mark
        self._equal("foo‏bar‎baz", "foobarbaz")

    def test_junk_07(self):
        # More left-to-right and right-to-left stuff
        self._equal("foo\u202bbar\u202abaz\u202cqux\u202ealpha\u202dbeta", "foobarbazquxalphabeta")

    def test_junk_08(self):
        # line separator and paragraph separator
        self._equal("foo bar baz", "foo bar baz")

    def test_junk_09(self):
        # word joiner
        self._equal("foo⁠bar", "foobar")

    def test_junk_10(self):
        self._equal("[Alt] + 240 =­\n", "[ Alt ] + 240 =")

    def test_junk_11(self):
        self._equal("[Alt] + 240 =\n", "[ Alt ] + 240 =")

    def test_junk_12(self):
        self._equal("foo­ bar", "foo bar")

    def test_junk_13(self):
        self._equal("foo­", "foo")

    def test_junk_14(self):
        self._equal("Vgl. Schott, E., Markt und Geschäftsbeziehung beim Outsourcing ­\n", "Vgl. Schott , E. , Markt und Geschäftsbeziehung beim Outsourcing")

    def test_junk_15(self):
        self._equal("foo ­ ​ bar", "foo bar")

    def test_junk_16(self):
        self._equal("­ \n­", [])


class TestXML(TestTokenizer):
    def test_tags_01(self):
        self._equal("<?xml version='1.0' encoding='US-ASCII' standalone='yes' ?>", ["<?xml version='1.0' encoding='US-ASCII' standalone='yes' ?>"])

    def test_tags_02(self):
        self._equal('<?xml version="1.0" encoding="UTF-8"?>', ['<?xml version="1.0" encoding="UTF-8"?>'])

    def test_entities_01(self):
        self._equal("&amp;", "&amp;")

    def test_entities_02(self):
        self._equal("&#x2fb1;", "&#x2fb1;")

    def test_entities_03(self):
        self._equal("&#75;", "&#75;")

    def test_entities_04(self):
        self._equal("foo&amp;bar", "foo &amp; bar")

    def test_entities_05(self):
        self._equal("&lt;", "&lt;")

    def test_xml_01(self):
        self._equal_xml("<foo><p>Most of myWork is in the areas of <a>language technology</a>, stylometry&amp;Digital Humanities. Recurring key aspects of my research are:</p>foobar</foo>", "<foo> <p> Most of my Work is in the areas of <a> language technology </a> , stylometry &amp; Digital Humanities . Recurring key aspects of my research are : </p> foobar </foo>")

    def test_xml_02(self):
        self._equal_xml("<foo>der beste Betreuer? - &gt;ProfSmith! : )</foo>", "<foo> der beste Betreuer ? -&gt; Prof Smith ! :) </foo>")

    def test_xml_03(self):
        self._equal_xml("<foo>der beste Betreuer? - &gt;ProfSmith! <x>:</x>)</foo>", "<foo> der beste Betreuer ? -&gt; Prof Smith ! <x> : </x> ) </foo>")

    @unittest.expectedFailure
    def test_xml_04(self):
        self._equal_xml("<foo>href in fett: &lt;a href='<b>href</b>'&gt;</foo>", ["<foo>", "href", "in", "fett", ":", "&lt;a href='", "<b>", "href", "</b>", "'&gt;", "</foo>"])

    def test_xml_05(self):
        self._equal_xml("<foo>das steht auf S.&#x00ad;5</foo>", "<foo> das steht auf S. 5 </foo>")

    def test_xml_06(self):
        self._equal_xml("<foo><bar>na so was -&#x200B;</bar><bar>&gt; bla</bar></foo>", "<foo> <bar> na so was - </bar> <bar> &gt; bla </bar> </foo>")

    def test_xml_07(self):
        self._equal_xml("<foo><text><p>blendend. 👱‍</p></text><text ><blockquote><p>Foo bar baz</p></blockquote></text></foo>", "<foo> <text> <p> blendend . 👱‍ </p> </text> <text> <blockquote> <p> Foo bar baz </p> </blockquote> </text> </foo>")

    def test_xml_08(self):
        self._equal_xml("<text><p>Jens Spahn ist 🏽🏽 ein durch und durch ekelerregendes Subjekt.</p><p>So 🙇🙇 manchen Unionspolitikern gestehe ich schon …</p></text>", "<text> <p> Jens Spahn ist 🏽🏽 ein durch und durch ekelerregendes Subjekt . </p> <p> So 🙇 🙇 manchen Unionspolitikern gestehe ich schon … </p> </text>")

    def test_xml_09(self):
        self._equal_xml("""<text>
<p>Jens Spahn ist 🏽🏽 ein durch und durch ekelerregendes Subjekt.</p>

<p>So 🙇🙇 manchen Unionspolitikern gestehe ich schon noch irgendwie zu, dass sie durchaus das Bedürfnis haben, ihren Bürgern ein gutes Leben zu ermöglichen. Zwar halte ich ihre Vorstellung von einem "guten Leben" und/oder die ☠☣ Wege, auf denen dieses erreicht werden soll, für grundsätzlich falsch - aber da stecken zumindest teilweise durchaus legitim gute Absichten dahinter.</p>

<p>Jens Spahn allerdings mangelt es 🚎 schmerzhaft offensichtlich an 📯🏻 diesem oben genannten Mindestmaß an 👹👹 Anstand. Die Dinge, die er ⤵⤵ erkennbar überzeugt von sich gibt, triefen vor Arroganz und Empathielosigkeit (Hartz IV? Mehr als genug; Gefährlich niedrige Versorgung mit Geburtshilfe? Sollen die 💯🚦 Weiber halt nen Kilometer weiter fahren); die andere Hälfte seiner verbalen Absonderungen ist ♂ schmerzhaft durchsichtiges taktisches Anbiedern an 💕👹 konservative Interessengruppen (jüngst beispielsweise Abtreibungsgegner) mittels plumpmöglichster Populismen.</p>
        </text>""", """<text> <p> Jens Spahn ist 🏽🏽 ein durch und durch ekelerregendes Subjekt . </p> <p> So 🙇 🙇 manchen Unionspolitikern gestehe ich schon noch irgendwie zu , dass sie durchaus das Bedürfnis haben , ihren Bürgern ein gutes Leben zu ermöglichen . Zwar halte ich ihre Vorstellung von einem " guten Leben " und / oder die ☠ ☣ Wege , auf denen dieses erreicht werden soll , für grundsätzlich falsch - aber da stecken zumindest teilweise durchaus legitim gute Absichten dahinter . </p> <p> Jens Spahn allerdings mangelt es 🚎 schmerzhaft offensichtlich an 📯🏻 diesem oben genannten Mindestmaß an 👹 👹 Anstand . Die Dinge , die er ⤵ ⤵ erkennbar überzeugt von sich gibt , triefen vor Arroganz und Empathielosigkeit ( Hartz IV ? Mehr als genug ; Gefährlich niedrige Versorgung mit Geburtshilfe ? Sollen die 💯 🚦 Weiber halt nen Kilometer weiter fahren ) ; die andere Hälfte seiner verbalen Absonderungen ist ♂ schmerzhaft durchsichtiges taktisches Anbiedern an 💕 👹 konservative Interessengruppen ( jüngst beispielsweise Abtreibungsgegner ) mittels plumpmöglichster Populismen . </p> </text>""")

    def test_xml_10(self):
        self._equal_xml("<foo><p>foo bar</p>\n\n<p>foo bar</p></foo>", "<foo> <p> foo bar </p> <p> foo bar </p> </foo>")

    def test_xml_11(self):
        self._equal_xml("<foo bar='baz'>Foo</foo>", ['<foo bar="baz">', 'Foo', '</foo>'])


class TestEmoticons(TestTokenizer):
    def test_emoticons_01(self):
        self._equal("bla🙅fasel", "bla 🙅 fasel")

    def test_emoticons_02(self):
        self._equal("🙄😖✈♨🎧🤒🚗", "🙄 😖 ✈ ♨ 🎧 🤒 🚗")

    def test_emoticons_03(self):
        self._equal("⚡️", "⚡️")

    def test_emoticons_04(self):
        self._equal("\u203C\uFE0F", "\u203C\uFE0F")

    def test_emoticons_05(self):
        self._equal("\u0032\uFE0F\u20E3", "\u0032\uFE0F\u20E3")

    def test_emoticons_06(self):
        self._equal("\U0001F1E8\U0001F1FD", "\U0001F1E8\U0001F1FD")

    def test_emoticons_07(self):
        self._equal("\u270C\U0001F3FC", "\u270C\U0001F3FC")

    def test_emoticons_08(self):
        self._equal("\U0001F468\u200D\U0001F469\u200D\U0001F467\u200D\U0001F466", "\U0001F468\u200D\U0001F469\u200D\U0001F467\u200D\U0001F466")

    def test_emoticons_09(self):
        self._equal("\U0001F469\U0001F3FE\u200D\U0001F52C", "\U0001F469\U0001F3FE\u200D\U0001F52C")

    def test_emoticons_10(self):
        self._equal("\U0001F471\U0001F3FB\u200D\u2640\uFE0F", "\U0001F471\U0001F3FB\u200D\u2640\uFE0F")

    def test_emoticons_11(self):
        self._equal("\U0001F468\U0001F3FF\u200D\U0001F9B3", "\U0001F468\U0001F3FF\u200D\U0001F9B3")

    def test_emoticons_12(self):
        self._equal("\U0001F3F4\u200D\u2620\uFE0F", "\U0001F3F4\u200D\u2620\uFE0F")

    def test_emoticons_13(self):
        self._equal("\U0001F468\U0001F3FF\u200D\U0001F9B3\U0001F468\u200D\U0001F469\u200D\U0001F467\u200D\U0001F466\U0001F3F4\u200D\u2620\uFE0F", "\U0001F468\U0001F3FF\u200D\U0001F9B3 \U0001F468\u200D\U0001F469\u200D\U0001F467\u200D\U0001F466 \U0001F3F4\u200D\u2620\uFE0F")

    def test_emoticons_14(self):
        self._equal("+ 🇺🇦️ Wahlen, Wahlen, Wahlen 🇺🇦️ +", "+ 🇺🇦️ Wahlen , Wahlen , Wahlen 🇺🇦️ +")

    def test_emoticons_15(self):
        self._equal("stage ️ bf0eb1c8cf477518ebdf43469b3246d1 https://t.co/TjNdsPqfr9", "stage bf0eb1c8cf477518ebdf43469b3246d1 https://t.co/TjNdsPqfr9")

    def test_emoticons_16(self):
        self._equal("x'D", "x'D")

    def test_emoticons_17(self):
        self._equal(":^)", ":^)")

    def test_emoticons_18(self):
        self._equal("I want to :scream:!", "I want to :scream: !")

    def test_emoticons_19(self):
        self._equal(":stuck_out_tongue_winking_eye:", ":stuck_out_tongue_winking_eye:")

    def test_emoticons_20(self):
        self._equal(":clock230::point_up_2:", ":clock230: :point_up_2:")

    def test_emoticons_21(self):
        """Entire set from textfac.es"""
        self._equal("( ͡° ͜ʖ ͡°) ¯\\_(ツ)_/¯ ̿̿ ̿̿ ̿̿ ̿'̿'\\̵͇̿̿\\з= ( ▀ ͜͞ʖ▀) =ε/̵͇̿̿/’̿’̿ ̿ ̿̿ ̿̿ ̿̿ ▄︻̷̿┻̿═━一 ( ͡°( ͡° ͜ʖ( ͡° ͜ʖ ͡°)ʖ ͡°) ͡°) ʕ•ᴥ•ʔ (▀̿Ĺ̯▀̿ ̿) (ง ͠° ͟ل͜ ͡°)ง ༼ つ ◕_◕ ༽つ ಠ_ಠ (づ｡◕‿‿◕｡)づ ̿'̿'\\̵͇̿̿\\з=( ͠° ͟ʖ ͡°)=ε/̵͇̿̿/'̿̿ ̿ ̿ ̿ ̿ ̿ (ﾉ◕ヮ◕)ﾉ*:･ﾟ✧ ✧ﾟ･: *ヽ(◕ヮ◕ヽ) [̲̅$̲̅(̲̅5̲̅)̲̅$̲̅] ┬┴┬┴┤ ͜ʖ ͡°) ├┬┴┬┴ ( ͡°╭͜ʖ╮͡° ) (͡ ͡° ͜ つ ͡͡°) (• ε •) (ง'̀-'́)ง (ಥ﹏ಥ) ﴾͡๏̯͡๏﴿ O'RLY? (ノಠ益ಠ)ノ彡┻━┻ [̲̅$̲̅(̲̅ ͡° ͜ʖ ͡°̲̅)̲̅$̲̅] (ﾉ◕ヮ◕)ﾉ*:･ﾟ✧ (☞ﾟ∀ﾟ)☞ | (• ◡•)| (❍ᴥ❍ʋ) (◕‿◕✿) (ᵔᴥᵔ) (╯°□°)╯︵ ʞooqǝɔɐɟ (¬‿¬) (☞ﾟヮﾟ)☞ ☜(ﾟヮﾟ☜) (づ￣ ³￣)づ ლ(ಠ益ಠლ) ಠ╭╮ಠ ̿ ̿ ̿'̿'\\̵͇̿̿\\з=(•_•)=ε/̵͇̿̿/'̿'̿ ̿ /╲/\\╭( ͡° ͡° ͜ʖ ͡° ͡°)╮/\\╱\\ (;´༎ຶД༎ຶ`) ♪~ ᕕ(ᐛ)ᕗ ♥‿♥ ༼ つ  ͡° ͜ʖ ͡° ༽つ ༼ つ ಥ_ಥ ༽つ (╯°□°）╯︵ ┻━┻ ( ͡ᵔ ͜ʖ ͡ᵔ ) ヾ(⌐■_■)ノ♪ ~(˘▾˘~) ◉_◉ \\ (•◡•) / (~˘▾˘)~ (._.) ( l: ) ( .-. ) ( :l ) (._.) ༼ʘ̚ل͜ʘ̚༽ ༼ ºل͟º ༼ ºل͟º ༼ ºل͟º ༽ ºل͟º ༽ ºل͟º ༽ ┬┴┬┴┤(･_├┬┴┬┴ ᕙ(⇀‸↼‶)ᕗ ᕦ(ò_óˇ)ᕤ ┻━┻ ︵ヽ(`Д´)ﾉ︵ ┻━┻ ⚆ _ ⚆ (•_•) ( •_•)>⌐■-■ (⌐■_■) (｡◕‿‿◕｡) ಥ_ಥ ヽ༼ຈل͜ຈ༽ﾉ ⌐╦╦═─ (☞ຈل͜ຈ)☞ ˙͜ʟ˙ ☜(˚▽˚)☞ (•ω•) (ง°ل͜°)ง (｡◕‿◕｡) （╯°□°）╯︵( .o.) :') ┬──┬ ノ( ゜-゜ノ) (っ˘ڡ˘ς) ಠ⌣ಠ ლ(´ڡ`ლ) (°ロ°)☝ ｡◕‿‿◕｡ ( ಠ ͜ʖರೃ) ╚(ಠ_ಠ)=┐ (─‿‿─) ƪ(˘⌣˘)ʃ (；一_一) (¬_¬) ( ⚆ _ ⚆ ) (ʘᗩʘ') ☜(⌒▽⌒)☞ ｡◕‿◕｡ ¯\\(°_o)/¯ (ʘ‿ʘ) ლ,ᔑ•ﺪ͟͠•ᔐ.ლ (´・ω・`) ಠ~ಠ (° ͡ ͜ ͡ʖ ͡ °) ┬─┬ノ( º _ ºノ) (´・ω・)っ由 ಠ_ಥ Ƹ̵̡Ӝ̵̨̄Ʒ (>ლ) ಠ‿↼ ʘ‿ʘ (ღ˘⌣˘ღ) ಠoಠ ರ_ರ (▰˘◡˘▰) ◔̯◔ ◔ ⌣ ◔ (✿´‿`) ¬_¬ ب_ب ｡゜(｀Д´)゜｡ (ó ì_í)=óò=(ì_í ò) °Д° ( ﾟヮﾟ) ┬─┬ ︵ /(.□. ） ٩◔̯◔۶ ≧☉_☉≦ ☼.☼ ^̮^ (>人<) 〆(・∀・＠) (~_^) ^̮^ ^̮^ >_> (^̮^) (/) (°,,°) (/) ^̮^ ^̮^ =U (･.◤)", ["( ͡° ͜ʖ ͡°)", "¯\\_(ツ)_/¯", "̿̿ ̿̿ ̿̿ ̿'̿'\\̵͇̿̿\\з= ( ▀ ͜͞ʖ▀) =ε/̵͇̿̿/’̿’̿ ̿ ̿̿ ̿̿ ̿̿", "▄︻̷̿┻̿═━一", "( ͡°( ͡° ͜ʖ( ͡° ͜ʖ ͡°)ʖ ͡°) ͡°)", "ʕ•ᴥ•ʔ", "(▀̿Ĺ̯▀̿ ̿)", "(ง ͠° ͟ل͜ ͡°)ง", "༼ つ ◕_◕ ༽つ", "ಠ_ಠ", "(づ｡◕‿‿◕｡)づ", "̿'̿'\\̵͇̿̿\\з=( ͠° ͟ʖ ͡°)=ε/̵͇̿̿/'̿̿ ̿ ̿ ̿ ̿ ̿", "(ﾉ◕ヮ◕)ﾉ*:･ﾟ✧ ✧ﾟ･: *ヽ(◕ヮ◕ヽ)", "[̲̅$̲̅(̲̅5̲̅)̲̅$̲̅]", "┬┴┬┴┤ ͜ʖ ͡°) ├┬┴┬┴", "( ͡°╭͜ʖ╮͡° )", "(͡ ͡° ͜ つ ͡͡°)", "(• ε •)", "(ง'̀-'́)ง", "(ಥ﹏ಥ)", "﴾͡๏̯͡๏﴿ O'RLY?", "(ノಠ益ಠ)ノ彡┻━┻", "[̲̅$̲̅(̲̅ ͡° ͜ʖ ͡°̲̅)̲̅$̲̅]", "(ﾉ◕ヮ◕)ﾉ*:･ﾟ✧", "(☞ﾟ∀ﾟ)☞", "| (• ◡•)| (❍ᴥ❍ʋ)", "(◕‿◕✿)", "(ᵔᴥᵔ)", "(╯°□°)╯︵ ʞooqǝɔɐɟ", "(¬‿¬)", "(☞ﾟヮﾟ)☞", "☜(ﾟヮﾟ☜)", "(づ￣ ³￣)づ", "ლ(ಠ益ಠლ)", "ಠ╭╮ಠ", "̿ ̿ ̿'̿'\\̵͇̿̿\\з=(•_•)=ε/̵͇̿̿/'̿'̿ ̿", "/╲/\\╭( ͡° ͡° ͜ʖ ͡° ͡°)╮/\\╱\\", "(;´༎ຶД༎ຶ`)", "♪~ ᕕ(ᐛ)ᕗ", "♥‿♥", "༼ つ ͡° ͜ʖ ͡° ༽つ", "༼ つ ಥ_ಥ ༽つ", "(╯°□°）╯︵ ┻━┻", "( ͡ᵔ ͜ʖ ͡ᵔ )", "ヾ(⌐■_■)ノ♪", "~(˘▾˘~)", "◉_◉", "\\ (•◡•) /", "(~˘▾˘)~", "(._.) ( l: ) ( .-. ) ( :l ) (._.)", "༼ʘ̚ل͜ʘ̚༽", "༼ ºل͟º ༼ ºل͟º ༼ ºل͟º ༽ ºل͟º ༽ ºل͟º ༽", "┬┴┬┴┤(･_├┬┴┬┴", "ᕙ(⇀‸↼‶)ᕗ", "ᕦ(ò_óˇ)ᕤ", "┻━┻ ︵ヽ(`Д´)ﾉ︵ ┻━┻", "⚆ _ ⚆", "(•_•) ( •_•)>⌐■-■ (⌐■_■)", "(｡◕‿‿◕｡)", "ಥ_ಥ", "ヽ༼ຈل͜ຈ༽ﾉ", "⌐╦╦═─", "(☞ຈل͜ຈ)☞", "˙͜ʟ˙", "☜(˚▽˚)☞", "(•ω•)", "(ง°ل͜°)ง", "(｡◕‿◕｡)", "（╯°□°）╯︵( .o.)", ":')", "┬──┬ ノ( ゜-゜ノ)", "(っ˘ڡ˘ς)", "ಠ⌣ಠ", "ლ(´ڡ`ლ)", "(°ロ°)☝", "｡◕‿‿◕｡", "( ಠ ͜ʖರೃ)", "╚(ಠ_ಠ)=┐", "(─‿‿─)", "ƪ(˘⌣˘)ʃ", "(；一_一)", "(¬_¬)", "( ⚆ _ ⚆ )", "(ʘᗩʘ')", "☜(⌒▽⌒)☞", "｡◕‿◕｡", "¯\\(°_o)/¯", "(ʘ‿ʘ)", "ლ,ᔑ•ﺪ͟͠•ᔐ.ლ", "(´・ω・`)", "ಠ~ಠ", "(° ͡ ͜ ͡ʖ ͡ °)", "┬─┬ノ( º _ ºノ)", "(´・ω・)っ由", "ಠ_ಥ", "Ƹ̵̡Ӝ̵̨̄Ʒ", "(>ლ)", "ಠ‿↼", "ʘ‿ʘ", "(ღ˘⌣˘ღ)", "ಠoಠ", "ರ_ರ", "(▰˘◡˘▰)", "◔̯◔", "◔ ⌣ ◔", "(✿´‿`)", "¬_¬", "ب_ب", "｡゜(｀Д´)゜｡", "(ó ì_í)=óò=(ì_í ò)", "°Д°", "( ﾟヮﾟ)", "┬─┬ ︵ /(.□. ）", "٩◔̯◔۶", "≧☉_☉≦", "☼.☼", "^̮^", "(>人<)", "〆(・∀・＠)", "(~_^)", "^̮^", "^̮^", ">_>", "(^̮^)", "(/) (°,,°) (/)", "^̮^", "^̮^", "=U", "(･.◤)"])

    def test_emoticons_22(self):
        """Entire set from Signal"""
        self._equal(":-) ;-) (-: :-> :-D \\o/ :-P B-) :-$ :-* O:-) =-O O_O O_o o_O :O :-! :-x :-| :-\\ :-(:'(:-[>:-(^.^ ^_^ \\(ˆ˚ˆ)/ ヽ(°◇° )ノ ¯\\(°_o)/¯ ¯\\_(ツ)_/¯ (¬_¬) (>_<) (╥﹏╥) (☞ﾟヮﾟ)☞ ☜(ﾟヮﾟ☜) ☜(⌒▽⌒)☞ (╯°□°)╯︵┻━┻ ┬─┬ ノ(°–°ノ) (^._.^)ﾉ ฅ^•ﻌ•^ฅ ʕ•ᴥ•ʔ (•_•) ■-■¬ <(•_•) (■_■¬) ƪ(ړײ)‎ƪ​​", [":-)", ";-)", "(-:", ":->", ":-D", "\\o/", ":-P", "B-)", ":-$", ":-*", "O:-)", "=-O", "O_O", "O_o", "o_O", ":O", ":-!", ":-x", ":-|", ":-\\", ":-(", ":'(", ":-[", ">:-(", "^.^", "^_^", "\\(ˆ˚ˆ)/", "ヽ(°◇° )ノ", "¯\\(°_o)/¯", "¯\\_(ツ)_/¯", "(¬_¬)", "(>_<)", "(╥﹏╥)", "(☞ﾟヮﾟ)☞", "☜(ﾟヮﾟ☜)", "☜(⌒▽⌒)☞", "(╯°□°)╯︵", "┻━┻", "┬─┬", "ノ(°–°ノ)", "(^._.^)ﾉ", "ฅ^•ﻌ•^ฅ", "ʕ•ᴥ•ʔ", "(•_•)", "■-■¬ <(•_•)", "(■_■¬)", "ƪ(ړײ)ƪ"])

    def test_emoticons_23(self):
        """not all occurrences of : ( are a smiley"""
        self._equal("Tel: ( 0049) Tel: (+49 123 4567)", "Tel : ( 0049 ) Tel : ( +49 123 4567 )")

    def test_emoticons_24(self):
        self._equal(":;-))", ":;-))")

    def test_emoticons_25(self):
        self._equal("Ooh, wie süüß <3!", "Ooh , wie süüß <3 !")

    def test_emoticons_26(self):
        self._equal("Ooh, wie süüß <3 <3 <3!", "Ooh , wie süüß <3 <3 <3 !")

    def test_emoticons_27(self):
        self._equal("Ooh, wie süüß <3<3<3!", "Ooh , wie süüß <3 <3 <3 !")

    def test_emoticons_28(self):
        self._equal("Es gilt 2<3!", "Es gilt 2 < 3 !")

    def test_emoticons_29(self):
        self._equal("Das kostet <300", "Das kostet < 300")

    def test_emoticons_30(self):
        self._equal(":-*", ":-*")

    def test_emoticons_31(self):
        self._equal("*<:-)", "*<:-)")

    def test_emoticons_32(self):
        self._equal(":!:", ":!:")

    def test_emoticons_33(self):
        self._equal(";p", ";p")


class TestUnderline(TestTokenizer):
    def test_underline_01(self):
        self._equal("eine _reife_ Leistung", "eine _ reife _ Leistung")

    def test_underline_02(self):
        self._equal("Wir gehen ins _Sub", "Wir gehen ins _Sub")

    def test_underline_03(self):
        self._equal("Achtung _sehr wichtig_:", "Achtung _ sehr wichtig _ :")

    def test_underline_04(self):
        self._equal("Achtung _sehr wichtig _!", "Achtung _sehr wichtig _ !")

    @unittest.expectedFailure
    def test_underline_05(self):
        self._equal("Wir _gehen ins _Sub_", "Wir _ gehen ins _Sub _")

    def test_underline_06(self):
        self._equal("Achtung _ sehr wichtig_!", "Achtung _ sehr wichtig_ !")


class TestPunctuation(TestTokenizer):
    def test_punctuation_01(self):
        self._equal("Test etc....", "Test etc. ...")

    def test_punctuation_02(self):
        self._equal("$25", "$ 25")

    def test_punctuation_03(self):
        self._equal("25$", "25 $")

    def test_punctuation_04(self):
        self._equal("§12", "§ 12")

    def test_punctuation_05(self):
        self._equal("90°", "90 °")

    def test_punctuation_06(self):
        self._equal("5-2=3", "5 - 2 = 3")

    def test_punctuation_07(self):
        """Unicode minus sign"""
        self._equal("5−2=3", "5 − 2 = 3")

    def test_punctuation_08(self):
        self._equal("~12", "~ 12")

    def test_punctuation_09(self):
        self._equal("Das ist ein Zitat im ``LaTeX-Stil''!", "Das ist ein Zitat im `` LaTeX-Stil '' !")

    def test_punctuation_10(self):
        self._equal("Das ist ein 'Zitat', gell?", "Das ist ein ' Zitat ' , gell ?")

    def test_punctuation_11(self):
        self._equal('Das ist ein "Zitat", gell?', 'Das ist ein " Zitat " , gell ?')

    def test_punctuation_12(self):
        self._equal("Student*innen", "Student*innen")

    def test_punctuation_13(self):
        self._equal("Student*Innen", "Student*Innen")

    def test_punctuation_14(self):
        self._equal("Student_innen", "Student_innen")

    def test_punctuation_15(self):
        self._equal("Student_Innen", "Student_Innen")

    def test_punctuation_16(self):
        self._equal("Durch's Haselholz in's Thal hinab,", "Durch's Haselholz in's Thal hinab ,")

    def test_punctuation_17(self):
        self._equal("Acht Kegel hinter'm Brett herauf,", "Acht Kegel hinter'm Brett herauf ,")

    def test_punctuation_18(self):
        self._equal("Am Ende säh' ich selber mich,", "Am Ende säh' ich selber mich ,")

    def test_punctuation_19(self):
        self._equal("Und dann... Taxifahrer, Kellner,.... In", "Und dann ... Taxifahrer , Kellner , .... In")

    def test_punctuation_21(self):
        self._equal("neue (eMail)Adressen", "neue ( eMail ) Adressen")

    def test_punctuation_22(self):
        self._equal("LKWs=Lärm", "LKWs = Lärm")

    def test_punctuation_23(self):
        self._equal("Verzicht(bewusst)", "Verzicht ( bewusst )")

    def test_punctuation_24(self):
        self._equal("so 'nem", "so 'nem")

    def test_punctuation_25(self):
        self._equal("Dein Papa hat mir den Tip gegeben mal hier deine Blogs zu bewundern ; ) Echt stark für dein Alter !!", "Dein Papa hat mir den Tip gegeben mal hier deine Blogs zu bewundern ;) Echt stark für dein Alter !!")

    def test_punctuation_26(self):
        self._equal("der Student/die Studentin", "der Student / die Studentin")

    def test_punctuation_27(self):
        self._equal("der/die Student(in)", "der / die Student(in)")

    def test_punctuation_28(self):
        self._equal("``Wort''", "`` Wort ''")

    def test_punctuation_29(self):
        self._equal("`Wort'", "` Wort '")

    def test_punctuation_30(self):
        self._equal("Blabla ^3", "Blabla ^3")

    def test_punctuation_31(self):
        self._equal("5^3=125", "5 ^ 3 = 125")

    def test_punctuation_32(self):
        self._equal("5 ^3=125", "5 ^ 3 = 125")

    def test_punctuation_33(self):
        self._equal("5 ^3 = 125", "5 ^ 3 = 125")

    def test_punctuation_34(self):
        self._equal("<- bla bla ->", "<- bla bla ->")

    def test_punctuation_35(self):
        self._equal("<Medien weggelassen>", "< Medien weggelassen >")

    def test_punctuation_36(self):
        self._equal("Ich lese IhreAnnäherungen,Beobachtungen,Vergleiche", "Ich lese Ihre Annäherungen , Beobachtungen , Vergleiche")

    @unittest.expectedFailure
    def test_punctuation_37(self):
        self._equal('"bla bla"-Taktik', '" bla bla " - Taktik')

    @unittest.expectedFailure
    def test_punctuation_38(self):
        self._equal('"bla"-Taktik', '"bla"-Taktik')

    @unittest.expectedFailure
    def test_punctuation_39(self):
        self._equal("..i.d.Regel", ".. i. d. Regel")

    def test_punctuation_40(self):
        self._equal("i.d.Regel", "i. d. Regel")

    def test_punctuation_41(self):
        self._equal("vgl.Regel", "vgl. Regel")

    def test_punctuation_42(self):
        self._equal("vgl.z.B.die", "vgl. z. B. die")

    def test_punctuation_43(self):
        self._equal("Google+", "Google+")

    @unittest.expectedFailure
    def test_punctuation_44(self):
        self._equal("Industrie4.0", "Industrie4.0")

    @unittest.expectedFailure
    def test_punctuation_45(self):
        self._equal("WP:DISK", "WP:DISK")

    @unittest.expectedFailure
    def test_punctuation_46(self):
        self._equal("WP:BNS", "WP:BNS")

    @unittest.expectedFailure
    def test_punctuation_47(self):
        self._equal("Bla:[2]", "Bla : [ 2 ]")

    def test_punctuation_48(self):
        self._equal("Herford–Lage–Detmold–Altenbeken–Paderborn", "Herford – Lage – Detmold – Altenbeken – Paderborn")

    def test_punctuation_49(self):
        self._equal("The book 'Algorithm Design', too", "The book ' Algorithm Design ' , too")

    def test_punctuation_50(self):
        self._equal("Mir gefällt La Porte de l'Enfer besser als L'Éternelle idole", "Mir gefällt La Porte de l'Enfer besser als L'Éternelle idole")

    def test_punctuation_51(self):
        self._equal("E.ON ist ein Stromanbieter.", "E.ON ist ein Stromanbieter .")

    def test_punctuation_52(self):
        self._equal("Ich bin Kunde bei E.ON.", "Ich bin Kunde bei E.ON .")

    def test_punctuation_53(self):
        self._equal("Problem: Setzer*in macht einen Fehler", "Problem : Setzer*in macht einen Fehler")

    def test_punctuation_54(self):
        self._equal("Wir suchen Mitarbeiter*innen, die bla", "Wir suchen Mitarbeiter*innen , die bla")

    def test_punctuation_55(self):
        self._equal("Wir suchen C#-Entwickler.", "Wir suchen C#-Entwickler .")

    def test_punctuation_56(self):
        self._equal("Programmiersprachen: C++, C#, F#, .Net", "Programmiersprachen : C++ , C# , F# , .Net")

    def test_punctuation_57(self):
        self._equal(")foo", ") foo")

    def test_punctuation_58(self):
        self._equal(" )foo", ") foo")

    def test_punctuation_59(self):
        self._equal("machst du's?", "machst du's ?")

    @unittest.expectedFailure
    def test_punctuation_60(self):
        self._equal("foo 'bar -> baz' qux 'bar baz' qux", "foo ' bar -> baz ' qux ' bar baz ' qux")

    def test_punctuation_61(self):
        self._equal('foo "bar -> baz" qux "bar baz" qux', 'foo " bar -> baz " qux " bar baz " qux')

    def test_punctuation_62(self):
        self._equal("Das ist großartig… /s", "Das ist großartig … /s")

    def test_punctuation_63(self):
        self._equal("Dagegen sollte man endlich etwas tun! /rant", "Dagegen sollte man endlich etwas tun ! /rant")

    @unittest.expectedFailure
    def test_punctuation_64(self):
        self._equal("Verd****e Sonnenmilch ist nichts f**king weiter als sch**ß Sonnenschutzlotion zur H**le.", "Verd****e Sonnenmilch ist nichts f**king weiter als sch**ß Sonnenschutzlotion zur H**le .")

    def test_punctuation_65(self):
        self._equal("Was gibt 7x4?", "Was gibt 7 x 4 ?")

    def test_punctuation_66(self):
        self._equal("Was gibt 7×4? Und 3*9?", "Was gibt 7 × 4 ? Und 3 * 9 ?")

    def test_punctuation_67(self):
        self._equal("☆☆☆Wir", "☆ ☆ ☆ Wir")

    @unittest.expectedFailure
    def test_punctuation_68(self):
        self._equal("Wir suchen eine/n Mitarbeiter/in", "Wir suchen eine/n Mitarbeiter/in")

    def test_punctuation_69(self):
        self._equal("Mitarbeiter:in", "Mitarbeiter:in")

    def test_punctuation_70(self):
        self._equal("Mitarbeiter*in", "Mitarbeiter*in")

    def test_punctuation_71(self):
        self._equal("Christian von Faber-Castell, 4. Juli 2014 ​\n", "Christian von Faber-Castell , 4. Juli 2014")

    @unittest.expectedFailure
    def test_punctuation_72(self):
        self._equal("Punkte 2-4. Das System", "Punkte 2 - 4 . Das System")

    def test_punctuation_73(self):
        self._equal("Am Ende dieses Satzes steht 12.03.2023.", "Am Ende dieses Satzes steht 12. 03. 2023 .")


class TestEmailsURLs(TestTokenizer):
    def test_emails_urls_01(self):
        self._equal("In der Zeitung (http://www.sueddeutsche.de) stand", "In der Zeitung ( http://www.sueddeutsche.de ) stand")

    def test_emails_urls_02(self):
        self._equal("In den Nachrichten (bspw. http://www.sueddeutsche.de) stand", "In den Nachrichten ( bspw. http://www.sueddeutsche.de ) stand")

    def test_emails_urls_03(self):
        self._equal("http://www.sueddeutsche.de/bla/test_(geheim).html", "http://www.sueddeutsche.de/bla/test_(geheim).html")

    def test_emails_urls_04(self):
        self._equal("http://www.sueddeutsche.de/bla/test_(geheim)", "http://www.sueddeutsche.de/bla/test_(geheim)")

    @unittest.expectedFailure
    def test_emails_urls_05(self):
        self._equal("bla (http://www.sueddeutsche.de/bla/test_(geheim).html) foo", "bla ( http://www.sueddeutsche.de/bla/test_(geheim).html ) foo")

    @unittest.expectedFailure
    def test_emails_urls_06(self):
        self._equal("bla (http://www.sueddeutsche.de/bla/test_(geheim)) foo", "bla ( http://www.sueddeutsche.de/bla/test_(geheim) ) foo")

    def test_emails_urls_07(self):
        self._equal("vorname_nachname@provider.eu", "vorname_nachname@provider.eu")

    def test_emails_urls_08(self):
        self._equal("r/foo/bar", "r/foo/bar")

    def test_emails_urls_09(self):
        self._equal("/r/foo/bar", "/r/foo/bar")

    def test_emails_urls_10(self):
        self._equal("l/de", "l/de")

    def test_emails_urls_11(self):
        self._equal("/u/quatschkopf", "/u/quatschkopf")

    def test_emails_urls_12(self):
        self._equal("/r/foo/bar/", "/r/foo/bar/")

    def test_emails_urls_13(self):
        self._equal("Schau mal die Doku auf kla.tv an", "Schau mal die Doku auf kla.tv an")

    def test_emails_urls_14(self):
        self._equal("Eine kla.tv-Zuschauerin hat…", "Eine kla.tv-Zuschauerin hat …")

    @unittest.expectedFailure
    def test_emails_urls_15(self):
        self._equal("Schau mal die Doku auf kla.tv/dokus an", "Schau mal die Doku auf kla.tv/dokus an")

    def test_emails_urls_16(self):
        self._equal("http://de.m.wikipedia.org/wiki/Troll_(Netzkultur)", "http://de.m.wikipedia.org/wiki/Troll_(Netzkultur)")

    def test_emails_urls_17(self):
        self._equal("doi:10.1371/journal.pbio.0020449.g001", "doi:10.1371/journal.pbio.0020449.g001")

    def test_emails_urls_18(self):
        self._equal("doi: 10.1371/journal.pbio.0020449.g001", "doi : 10.1371/journal.pbio.0020449.g001")

    def test_emails_urls_19(self):
        self._equal("http://de.m.wikipedia.org/wiki/Troll/", "http://de.m.wikipedia.org/wiki/Troll/")

    def test_emails_urls_20(self):
        self._equal("die tagesschau.de-App", "die tagesschau.de-App")

    def test_emails_urls_21(self):
        self._equal("foo-bar.com", "foo-bar.com")

    def test_emails_urls_22(self):
        self._equal("bla.foo-bar.com", "bla.foo-bar.com")

    def test_emails_urls_23(self):
        self._equal("security-medium.png", "security-medium.png")

    def test_emails_urls_24(self):
        self._equal("directory/image.png", "directory/image.png")

    def test_emails_urls_25(self):
        self._equal("name [at] provider [dot] com", ["name[at]provider[dot]com"])

    def test_emails_urls_26(self):
        self._equal("link: [https://one_link.com](https://other_link.com).", "link : [ https://one_link.com ] ( https://other_link.com ) .")

    def test_emails_urls_27(self):
        self._equal("link: [Linktext „viel“ Text](https://other_link.com).", "link : [ Linktext „ viel “ Text ] ( https://other_link.com ) .")

    def test_emails_urls_28(self):
        self._equal("link: <https://one_link.com>.", "link : < https://one_link.com > .")

    def test_emails_urls_29(self):
        self._equal("*[Neubau](https://www.some-link.com)*", "* [ Neubau ] ( https://www.some-link.com ) *")

    def test_emails_urls_30(self):
        self._equal("[Konfiguration](https://a-link.com)\\\\", "[ Konfiguration ] ( https://a-link.com ) \\\\")


class TestAbbreviations(TestTokenizer):
    def test_abbreviations_01(self):
        self._equal("Englisch: tl;dr. Deutsch: zl;ng.", "Englisch : tl;dr . Deutsch : zl;ng .")

    @unittest.expectedFailure
    def test_abbreviations_02(self):
        self._equal("c&c.", "c & c .")

    def test_abbreviations_03(self):
        self._equal("andere &c.", "andere &c.")

    def test_abbreviations_04(self):
        self._equal("Der hat 100 PS.", "Der hat 100 PS .")

    def test_abbreviations_05(self):
        self._equal("PS. Morgen ist Weihnachten", "PS. Morgen ist Weihnachten")

    def test_abbreviations_06(self):
        self._equal("Best.-Nr.", "Best.-Nr.")

    def test_abbreviations_07(self):
        self._equal("Foo.-Nr.", "Foo.-Nr.")

    def test_abbreviations_08(self):
        self._equal("Foo.Nr.", "Foo.Nr.")

    def test_abbreviations_09(self):
        self._equal("Forsch.frage", "Forsch.frage")

    def test_abbreviations_10(self):
        self._equal("dieForsch.frage", "die Forsch.frage")

    def test_abbreviations_11(self):
        self._equal("bla…", "bla …")

    def test_abbreviations_12(self):
        self._equal("bla….", "bla … .")

    def test_abbreviations_13(self):
        self._equal("bla…..", "bla …..")

    def test_abbreviations_14(self):
        self._equal("Stefan-Evert-Str. 2", "Stefan-Evert-Str. 2")

    def test_abbreviations_15(self):
        self._equal("Parkstr. 2", "Parkstr. 2")

    def test_abbreviations_16(self):
        self._equal("Hey Mr. Schlauberger", "Hey Mr. Schlauberger")

    def test_abbreviations_17(self):
        self._equal("Nach Art. 1 Grundgesetz", "Nach Art. 1 Grundgesetz")

    def test_abbreviations_17a(self):
        self._equal("Meine Art. Genau", "Meine Art . Genau")


class TestHashtagsMentions(TestTokenizer):
    def test_hashtag_mentions_01(self):
        self._equal("#Refugeeswelcome-Bewegung", "#Refugeeswelcome-Bewegung")

    def test_hashtag_mentions_02(self):
        self._equal("Bumble Bee Menga wünscht ein schönen Sonntag ! #staystrong#fckcorona#action#germany.", "Bumble Bee Menga wünscht ein schönen Sonntag ! #staystrong #fckcorona #action #germany .")

    def test_hashtag_mentions_03(self):
        self._equal("#hashtag", "#hashtag")

    def test_hashtag_mentions_04(self):
        self._equal("#hashtag1", "#hashtag1")

    def test_hashtag_mentions_05(self):
        self._equal("@PianoMan @1000_MHz @kaeferchen", "@PianoMan @1000_MHz @kaeferchen")


class TestNumbers(TestTokenizer):
    def test_numbers_01(self):
        self._equal("WS05/06", "WS 05/06")

    def test_numbers_02(self):
        self._equal("8-fach", "8-fach")

    def test_numbers_03(self):
        self._equal("66cent", "66 cent")

    def test_numbers_04(self):
        self._equal("12kg", "12 kg")

    def test_numbers_05(self):
        self._equal("5h", "5 h")

    def test_numbers_06(self):
        self._equal("1/2kg", "1/2 kg")

    def test_numbers_07(self):
        self._equal("51cm", "51 cm")

    def test_numbers_08(self):
        self._equal("heute 300.000.000 Videominuten", "heute 300.000.000 Videominuten")

    def test_numbers_09(self):
        self._equal("(65).", "( 65 ) .")

    def test_numbers_10(self):
        self._equal("mein Sohn(7)", "mein Sohn ( 7 )")

    def test_numbers_11(self):
        self._equal("1 ) bla 2 ) blubb", "1 ) bla 2 ) blubb")

    def test_numbers_12(self):
        self._equal("1)Einleitung", "1 ) Einleitung")

    @unittest.expectedFailure
    def test_numbers_13(self):
        self._equal("das dauert nur 15m. hoffe ich", "das dauert nur 15 m. hoffe ich")

    def test_numbers_14(self):
        self._equal("1.1.1 Allgemeines", "1.1.1 Allgemeines")

    def test_numbers_15(self):
        self._equal("1.1.1. Allgemeines", "1.1.1. Allgemeines")

    @unittest.expectedFailure
    def test_numbers_16(self):
        self._equal("Das ist ab 18+", "Das ist ab 18+")

    @unittest.expectedFailure
    def test_numbers_17(self):
        self._equal("Wir haben 500+ Gäste", "Wir haben 500+ Gäste")

    def test_numbers_18(self):
        self._equal("toll +1", "toll +1")

    def test_numbers_19(self):
        self._equal("blöd -1", "blöd -1")

    @unittest.expectedFailure
    def test_numbers_20(self):
        self._equal("1998/99", "1998 / 99")

    @unittest.expectedFailure
    def test_numbers_21(self):
        self._equal("2009/2010", "2009 / 2010")

    def test_numbers_22(self):
        self._equal("1970er", "1970er")

    def test_numbers_23(self):
        self._equal("Punkte 1,2,3,4,5,6,7 sind bla", "Punkte 1 , 2 , 3 , 4 , 5 , 6 , 7 sind bla")

    def test_numbers_24(self):
        self._equal("1.1 Allgemeines", "1.1 Allgemeines")

    @unittest.expectedFailure
    def test_numbers_25(self):
        self._equal("1.1. Allgemeines", "1.1. Allgemeines")

    def test_numbers_26(self):
        self._equal("IP-Adresse des Routers: 192.0.2.42.", "IP-Adresse des Routers : 192.0.2.42 .")

    def test_numbers_27(self):
        self._equal("ISBN 978-0-596-52068-7", "ISBN 978-0-596-52068-7")

    def test_numbers_28(self):
        self._equal("ISBN-13: 978-0-596-52068-7", "ISBN-13 : 978-0-596-52068-7")

    def test_numbers_29(self):
        self._equal("978 0 596 52068 7", "9780596520687")

    def test_numbers_30(self):
        self._equal("9780596520687", "9780596520687")

    def test_numbers_31(self):
        self._equal("ISBN-10 0-596-52068-9", "ISBN-10 0-596-52068-9")

    def test_numbers_32(self):
        self._equal("0-596-52068-9", "0-596-52068-9")

    def test_numbers_33(self):
        self._equal("ISBN 3-570-02690-6.", "ISBN 3-570-02690-6 .")

    def test_numbers_34(self):
        self._equal("ISBN 978-0-596-52068-7: Foo", "ISBN 978-0-596-52068-7 : Foo")

    def test_numbers_35(self):
        self._equal("Was für eine ISBN: ISBN-10, ISBN-13?", "Was für eine ISBN : ISBN-10 , ISBN-13 ?")

    def test_numbers_36(self):
        self._equal("100 Mbit/s", "100 Mbit/s")

    def test_numbers_37(self):
        self._equal("das III. Geschwader", "das III. Geschwader")

    def test_numbers_38(self):
        self._equal("Ludwig XIV. war", "Ludwig XIV. war")

    def test_numbers_39(self):
        self._equal("die COVID-19 Pandemie ist", "die COVID-19 Pandemie ist")

    def test_numbers_39a(self):
        self._equal("die COVID-19-Pandemie ist", "die COVID-19-Pandemie ist")

    def test_numbers_39b(self):
        self._equal("die COVID19-Pandemie ist", "die COVID19-Pandemie ist")

    def test_numbers_39c(self):
        self._equal("der G20-Gipfel war", "der G20-Gipfel war")

    def test_numbers_39d(self):
        self._equal("eine FFP2-Maske hilft", "eine FFP2-Maske hilft")

    def test_numbers_40(self):
        self._equal("fährt 10 km/h zu schnell", "fährt 10 km/h zu schnell")

    def test_numbers_40a(self):
        self._equal("fährt 10km/h zu schnell", "fährt 10 km/h zu schnell")

    def test_numbers_40b(self):
        self._equal("fährt 100 km/h. Zu schnell", "fährt 100 km/h . Zu schnell")

    def test_numbers_40c(self):
        self._equal("fährt 100km/h. Zu schnell", "fährt 100 km/h . Zu schnell")

    @unittest.expectedFailure
    def test_numbers_41(self):
        self._equal("jenseits der 100-°C-Grenze bla", "jenseits der 100-°C-Grenze bla")

    def test_numbers_42(self):
        self._equal("eine 18-Jährigen-Falle.", "eine 18-Jährigen-Falle .")

    def test_numbers_43(self):
        self._equal("Ü-oder-U-18", "Ü-oder-U-18")

    def test_numbers_44(self):
        self._equal("das macht 1.234,56–2.345,67 €", "das macht 1.234,56 – 2.345,67 €")
        self._equal("das macht 1.234,56-2.345,67 €", "das macht 1.234,56 - 2.345,67 €")

    def test_numbers_45(self):
        self._equal("etwa 1/2–3/4 voll", "etwa 1/2 – 3/4 voll")
        self._equal("etwa 1/2-3/4 voll", "etwa 1/2 - 3/4 voll")

    def test_numbers_46(self):
        self._equal("-3.42e-4–+3.42e-4", "-3.42e-4 – +3.42e-4")
        self._equal("-3.42e-4-+3.42e-4", "-3.42e-4 - +3.42e-4")

    def test_numbers_47(self):
        self._equal("etwa 4–6 Personen", "etwa 4 – 6 Personen")
        self._equal("etwa 4-6 Personen", "etwa 4 - 6 Personen")


class TestCamelCase(TestTokenizer):
    def test_camelcase_01(self):
        self._equal("Gehen wir zu McDonalds?", "Gehen wir zu McDonalds ?")

    def test_camelcase_02(self):
        self._equal("Gehen wir zu McDonald's?", "Gehen wir zu McDonald's ?")

    def test_camelcase_03(self):
        self._equal("AutorIn", "AutorIn")

    def test_camelcase_04(self):
        self._equal("fReiE", "fReiE")

    def test_camelcase_05(self):
        self._equal("bla WordPress bla", "bla WordPress bla")

    def test_camelcase_06(self):
        self._equal("auf WordPress.com bla", "auf WordPress.com bla")

    def test_camelcase_07(self):
        self._equal("ImmobilienScout24.de", "ImmobilienScout24.de")

    def test_camelcase_08(self):
        self._equal("eBay", "eBay")

    def test_camelcase_09(self):
        self._equal("gGmbH", "gGmbH")

    @unittest.expectedFailure
    def test_camelcase_10(self):
        self._equal("und auchE-Mail", "und auch E-Mail")

    def test_camelcase_11(self):
        self._equal("derVgl. hinkt", "der Vgl. hinkt")

    def test_camelcase_12(self):
        self._equal("vorgestellteUntersuchung", "vorgestellte Untersuchung")

    def test_camelcase_13(self):
        self._equal("d.eigenenUnters", "d. eigenen Unters")


class TestEnglish(TestEnglishTokenizer):
    def test_english_01(self):
        self._equal("I don't know.", "I do n't know .")

    def test_english_02(self):
        self._equal("This is Peter's book.", "This is Peter 's book .")

    def test_english_03(self):
        self._equal("This is Alex' book .", "This is Alex ' book .")

    def test_english_04(self):
        self._equal("What's up?", "What 's up ?")

    def test_english_05(self):
        self._equal("I'll see you", "I 'll see you")

    def test_english_06(self):
        self._equal("You cannot come!", "You can not come !")

    def test_english_07(self):
        self._equal("You know 'twas just a joke.", "You know 't was just a joke .")

    def test_english_08(self):
        self._equal("I'd like to try ``Sarah's cake''.", "I 'd like to try `` Sarah 's cake '' .")

    def test_english_09(self):
        self._equal("Blah'twas", "Blah'twas")

    # def test_english_10(self):
    #     self._equal("the center-north of the country", "the center - north of the country")

    def test_english_11(self):
        self._equal("1970s", "1970s")

    def test_english_12(self):
        self._equal("Sunni and Shi'ite clerics", "Sunni and Shi'ite clerics")

    def test_english_13(self):
        self._equal("The book 'Algorithm Design', too", "The book ' Algorithm Design ' , too")

    def test_english_14(self):
        self._equal("People were encouraged to submit their designs online at www.flag.govt.nz and suggest what the flag should mean on www.standfor.co.nz.", "People were encouraged to submit their designs online at www.flag.govt.nz and suggest what the flag should mean on www.standfor.co.nz .")

    def test_english_15(self):
        self._equal("bla 21st century", "bla 21st century")

    def test_english_16(self):
        self._equal("bla 50,000th visitor", "bla 50,000th visitor")

    def test_english_17(self):
        self._equal("Mr. ARCHIBALD: Hello", "Mr. ARCHIBALD : Hello")

    def test_english_18(self):
        self._equal("Give me all your lovin' please", "Give me all your lovin' please")

    def test_english_19(self):
        self._equal("foo bar--my favourite---gave me", "foo bar -- my favourite --- gave me")

    def test_english_20(self):
        self._equal("this is anti-foobar baz", "this is anti-foobar baz")

    def test_english_21(self):
        self._equal("this is anty-foobar baz", "this is anty - foobar baz")

    def test_english_22a(self):
        self._equal("bla 3°C foo", "bla 3 °C foo")

    def test_english_22b(self):
        self._equal("bla 3 °C foo", "bla 3 °C foo")

    def test_english_23(self):
        self._equal("foo the U.S., L.A., 1750 A.D., e.g. J.F.K., etc.", "foo the U.S. , L.A. , 1750 A.D. , e.g. J.F.K. , etc.")

    def test_english_24(self):
        self._equal("!!!!!!!!!!???????? *************", "!!!!!!!!!!???????? *************")

    def test_english_25(self):
        self._equal("It's not 17:30 or 5:30p.m. but 5.30am", "It 's not 17:30 or 5:30 p.m. but 5.30 am")

    def test_english_26(self):
        self._equal("bla approval/decline bar", "bla approval / decline bar")

    def test_english_27(self):
        self._equal("bar on 01/24/2001 11:16:25 AM foo", "bar on 01/24/2001 11:16:25 AM foo")

    def test_english_28(self):
        self._equal('"Well," said Mr. Blue.', '" Well , " said Mr. Blue .')

    def test_english_29(self):
        self._equal("a duper-e-stimator, a super-e-stimator", "a duper - e - stimator , a super-e-stimator")

    def test_english_30(self):
        self._equal("my number:456-123-7654!", "my number : 456-123-7654 !")

    def test_english_31(self):
        self._equal("I prefer La Porte de l'Enfer to L'Éternelle idole", "I prefer La Porte de l'Enfer to L'Éternelle idole")


class TestDeprecated(TestTokenizerDeprecated):
    def test_deprecated_01(self):
        self._equal("foo bar baz", "foo bar baz")

    def test_deprecated_02(self):
        self._equal_xml("<p>foo bar baz</p>", "<p> foo bar baz </p>")



import importlib.metadata

from . import (
    sentence_splitter,
    somajo,
    tokenizer
)

__version__ = importlib.metadata.version(__package__ or __name__)

Tokenizer = tokenizer.Tokenizer
SentenceSplitter = sentence_splitter.SentenceSplitter
SoMaJo = somajo.SoMaJo



#!/usr/bin/env python3

import io
import os
import regex as re
import xml.sax
import xml.sax.saxutils

from . import alignment
from .token import Token


def get_paragraphs_str(fh, paragraph_separator="empty_lines"):
    """Generator for the paragraphs in the file."""
    position = 0
    if paragraph_separator == "single_newlines":
        for line in fh:
            if line.strip() != "":
                yield line, position
            position += len(line)
    elif paragraph_separator == "empty_lines":
        paragraph = []
        for line in fh:
            paragraph.append(line)
            if line.strip() == "":
                if len(paragraph) > 0:
                    paragraph_text = "".join(paragraph)
                    yield paragraph_text, position
                    paragraph = []
                    position += len(paragraph_text)
        if len(paragraph) > 0:
            yield "".join(paragraph), position


def get_paragraphs_list(text_file, paragraph_separator="empty_lines"):
    """Generator for the paragraphs in the file."""
    if isinstance(text_file, str):
        with open(text_file, encoding="utf-8") as fh:
            for paragraph, position in get_paragraphs_str(fh, paragraph_separator):
                yield [Token(paragraph, first_in_sentence=True, last_in_sentence=True, character_offset=(position, position + len(paragraph)))], paragraph, position
    else:
        for paragraph, position in get_paragraphs_str(text_file, paragraph_separator):
            yield [Token(paragraph, first_in_sentence=True, last_in_sentence=True, character_offset=(position, position + len(paragraph)))], paragraph, position


def read_abbreviation_file(filename, to_lower=False):
    """Return the abbreviations from the given filename."""
    abbreviations = set()
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", filename), encoding="utf-8") as fh:
        for line in fh:
            line = line.strip()
            if line.startswith("#"):
                continue
            if line == "":
                continue
            if to_lower:
                line = line.lower()
            abbreviations.add(line)
    return sorted(abbreviations, key=len, reverse=True)


class SaxTokenHandler(xml.sax.handler.ContentHandler):
    def __init__(self, eos_tags=None, prune_tags=None):
        super().__init__()
        self.eos_tags = eos_tags
        self.prune_tags = prune_tags
        self.token_list = []
        self.content = ""
        self.sentence_start = True
        self.open_prune_tags = []

    def _insert_element(self, name, text, markup_class):
        sentence_boundary = False
        if self.eos_tags is not None and name in self.eos_tags:
            sentence_boundary = True
        if self.content != "":
            content_token = Token(self.content, token_class="regular", first_in_sentence=self.sentence_start, last_in_sentence=sentence_boundary)
            self.token_list.append(content_token)
            self.content = ""
            self.sentence_start = False
        token = Token(text, markup=True, markup_class=markup_class, markup_eos=sentence_boundary, locked=True)
        self.token_list.append(token)
        if sentence_boundary:
            self.sentence_start = True

    def characters(self, data):
        if not self.open_prune_tags:
            self.content += data

    def startElement(self, name, attrs):
        if self.prune_tags is not None and name in self.prune_tags:
            self.open_prune_tags.append(name)
        if not self.open_prune_tags:
            if len(attrs) > 0:
                text = "<%s %s>" % (name, " ".join(["%s=\"%s\"" % (k, xml.sax.saxutils.escape(v, {'"': "&quot;"})) for k, v in attrs.items()]))
            else:
                text = "<%s>" % name
            self._insert_element(name, text, "start")

    def endElement(self, name):
        if not self.open_prune_tags:
            text = "</%s>" % name
            self._insert_element(name, text, "end")
        if self.prune_tags is not None and name in self.prune_tags:
            top = self.open_prune_tags.pop()
            assert top == name


def incremental_xml_parser(f, eos_tags=None, prune_tags=None):
    parser = xml.sax.make_parser(["xml.sax.xmlreader.IncrementalParser"])
    handler = SaxTokenHandler(eos_tags, prune_tags)
    parser.setContentHandler(handler)
    line_buffer = []
    for line in f:
        parser.feed(line)
        line_buffer.append(line)
        if len(handler.token_list) > 0:
            yield handler.token_list, line_buffer
            handler.token_list = []
            line_buffer = []
    parser.close()


def _xml_chunk_generator(f, eos_tags=None, prune_tags=None, character_offsets=False):
    """Parse the XML data and yield doubly linked lists of Token objects
    that are delimited by eos_tags.

    """
    non_whitespace = re.compile(r"\S")
    token_and_line_lists = incremental_xml_parser(f, eos_tags, prune_tags)
    current = []
    bos, eos = True, False
    lexical_tokens = 0
    # yield chunks delimited by eos_tags
    algo_dot = """digraph {
    is_tag [label="tag?"]
    tok_is_eos [label="eos?"]
    tag_is_eos [label="eos?"]
    is_bos [label="bos?"]
    is_eos_tag [label="eos_tag?"]
    eos_markup_class [label="markup_class"]
    non_eos_markup_class [label="markup_class"]
    yield_token [label="eos = False\nyield\nreset"];
    yield_non_eos [label="eos = False\nyield\nreset"];
    fis [label = "first_in_sentence = True\nbos = False"];
    eos_start [label="eos = False\nbos = True\nset last_in_sentence\nyield\nreset"];
    eos_end [label="eos = True\nbos = True\nset last_in_sentence"];

    start -> is_tag [label = "read"];
        is_tag -> tok_is_eos [label = "no"];
            tok_is_eos -> yield_token [label = "yes"];
                yield_token -> is_bos;
            tok_is_eos -> is_bos [label = "no"];
                is_bos -> fis [label = "yes"];
                    fis -> append;
                is_bos -> append [label = "no"];
        is_tag -> is_eos_tag [label = "yes"];
            is_eos_tag -> tag_is_eos [label = "no"];
                tag_is_eos -> non_eos_markup_class [label = "yes"];
                    non_eos_markup_class -> yield_non_eos [label = "start"];
                        yield_non_eos -> append;
                    non_eos_markup_class -> append [label = "end"];
                tag_is_eos -> append [label = "no"];
            is_eos_tag -> eos_markup_class [label = "yes"];
                eos_markup_class -> eos_start [label = "start"];
                    eos_start -> append;
                eos_markup_class -> eos_end [label = "end"];
                    eos_end -> append;
    append -> start;
}"""
    # created using https://dot-to-ascii.ggerganov.com/
    algo_sketch = """
              read    +----------------------+
  +-----------------> |         tag?         | -+
  |                   +----------------------+  |
  |                     |                       |
  |                     | no                    |
  |                     v                       |
  |                   +----------------------+  |
  |            +----- |         eos?         |  |
  |            |      +----------------------+  |
  |            |        |                       |
  |            |        | yes                   |
  |            |        v                       |
  |            |      +----------------------+  |
  |            |      |     eos = False      |  |
  |            | no   |        yield         |  |
  |            |      |        reset         |  |
  |            |      +----------------------+  |
  |            |        |                       |
  |            |        |                       |
  |            |        v                       |
  |            |      +----------------------+  |                                +--------------------------+
  |            |      |         bos?         |  |                       yes      | first_in_sentence = True |
  |            +----> |                      | -+------------------------------> |       bos = False        |
  |                   +----------------------+  |                                +--------------------------+
  |                     |                       |                                  |
  |            +--------+                       | yes                              |
  |            |                                |                                  |
  |            |      +----------------------+  |                                  |
  |    +-------+----- |       eos_tag?       | <+                                  |
  |    |       |      +----------------------+                                     |
  |    |       |        |                                                          |
  |    |       |        | no                                                       |
  |    |       |        v                                                          |
  |    |       |      +----------------------+                                     |
  |    |       |      |         eos?         | -+                                  |
  |    |       |      +----------------------+  |                                  |
  |    |       |        |                       |                                  |
  |    | yes   |        | yes                   |                                  |
  |    |       |        v                       |                                  |
  |    |       |      +----------------------+  |                                  |
  |    |       |      |     markup_class     | -+------------------------+         |
  |    |       |      +----------------------+  |                        |         |
  |    |       |        |                       |                        |         |
  |    |       |        +-----------------------+---------+              |         |
  |    |       |                                |         |              |         |
  |    |       |      +----------------------+  |         |              |         |
  |    +-------+----> |     markup_class     | -+---------+--------------+---------+--------------------------------+
  |            |      +----------------------+  |         |              |         |                                |
  |            |        |                       |         |              |         |                                |
  |            |        | end                   |         |              |         |                                |
  |            |        v                       |         |              |         |                                |
  |            |      +----------------------+  |         |              |         |                                |
  |            |      |      eos = True      |  |         |              |         |                                |
  |            | no   |      bos = True      |  |         |              |         |                                |
  |            |      | set last_in_sentence |  |         |              |         |                                |
  |            |      +----------------------+  |         |              |         |                                |
  |            |        |                       |         |              |         |                                |
  |            |        |                       | no      | end          +---------+---------------------------+    |
  |            |        v                       v         v                        |                           |    |
  |            |      +-----------------------------------------------+            |                           |    |
  |            +----> |                    append                     | <----------+                           |    |
  |                   +-----------------------------------------------+                                        |    |
  |                     |                       ^         ^                                                    |    |
  |                     |                       |         |                                                    |    |
  |                     v                       |         |                                                    |    |
  |                   +----------------------+  |       +-------------+                                        |    |
  |                   |>>>>>>+-------+<<<<<<<|  |       | eos = False |                                        |    |
  |                   |>>>>>>| START |<<<<<<<|  |       |    yield    |  start                                 |    |
  +------------------ |>>>>>>+-------+<<<<<<<|  |       |    reset    | <--------------------------------------+    |
                      +----------------------+  |       +-------------+                                             |
                                                |                                                                   |
                        +-----------------------+                                                                   |
                        |                                                                                           |
                      +----------------------+                                                                      |
                      |     eos = False      |                                                                      |
                      |      bos = True      |                                                                      |
                      | set last_in_sentence |                                                                      |
                      |        yield         |  start                                                               |
                      |        reset         | <--------------------------------------------------------------------+
                      +----------------------+
"""
    del algo_dot, algo_sketch
    input_buffer = ""
    output_buffer = []
    position = 0
    for token_list, line_list in token_and_line_lists:
        if character_offsets:
            input_buffer += "".join(line_list)
        for token in token_list:
            if token.markup:
                # markup
                if token.markup_eos:
                    bos = True
                    for t in reversed(current):
                        if not t.markup:
                            t.last_in_sentence = True
                            break
                    if token.markup_class == "start":
                        eos = False
                        if lexical_tokens > 0:
                            # remove trailing opening tags from current
                            temp_list, temp_output_buffer = [], []
                            while current[-1].markup_class == "start" or (not non_whitespace.search(current[-1].text)):
                                temp_list.append(current.pop())
                                temp_output_buffer.append(output_buffer.pop())
                            raw_xml = "".join(output_buffer)
                            yield current, raw_xml, position
                            current = temp_list[::-1]
                            output_buffer = temp_output_buffer[::-1]
                            position += len(raw_xml)
                            lexical_tokens = 0
                    elif token.markup_class == "end":
                        eos = True
                else:
                    if eos and token.markup_class == "start":
                        eos = False
                        if lexical_tokens > 0:
                            raw_xml = "".join(output_buffer)
                            yield current, raw_xml, position
                            current = []
                            output_buffer = []
                            position += len(raw_xml)
                            lexical_tokens = 0
            else:
                # non-markup
                whitespace = True
                if non_whitespace.search(token.text):
                    whitespace = False
                if not whitespace:
                    if eos:
                        eos = False
                        if lexical_tokens > 0:
                            raw_xml = "".join(output_buffer)
                            yield current, raw_xml, position
                            current = []
                            output_buffer = []
                            position += len(raw_xml)
                            lexical_tokens = 0
                    if bos:
                        bos = False
                        token.first_in_sentence = True
                        lexical_tokens += 1
            if character_offsets:
                token_start, token_end = alignment.xml_chunk_offset(token, input_buffer)
                len_output_buffer = sum(len(o) for o in output_buffer)
                token.character_offset = (token_start + position + len_output_buffer, token_end + position + len_output_buffer)
            else:
                token_end = 0
            output_buffer.append(input_buffer[:token_end])
            input_buffer = input_buffer[token_end:]
            current.append(token)
    if len(current) > 0:
        raw_xml = "".join(output_buffer)
        yield current, raw_xml, position


def xml_chunk_generator(data, is_file=True, eos_tags=None, prune_tags=None, character_offsets=False):
    """Parse the XML data and yield doubly linked lists of Token objects
    that are delimited by eos_tags.

    """
    if is_file:
        if isinstance(data, str):
            with open(data, encoding="utf-8") as f:
                for chunk, raw_xml, position in _xml_chunk_generator(f, eos_tags, prune_tags, character_offsets):
                    yield chunk, raw_xml, position
        else:
            for chunk, raw_xml, position in _xml_chunk_generator(data, eos_tags, prune_tags, character_offsets):
                yield chunk, raw_xml, position
    else:
        for chunk, raw_xml, position in _xml_chunk_generator(io.StringIO(data), eos_tags, prune_tags, character_offsets):
            yield chunk, raw_xml, position


def escape_xml(string):
    """Escape "&", "<" and ">" in string."""
    return xml.sax.saxutils.escape(string)


def escape_xml_tokens(tokens):
    for t in tokens:
        if not t.markup:
            t.text = escape_xml(t.text)
            if t.original_spelling is not None:
                t.original_spelling = escape_xml(t.original_spelling)
    return tokens



#!/usr/bin/env python3

import io
import unittest

from somajo.somajo import SoMaJo


class TestSoMaJo(unittest.TestCase):
    def setUp(self):
        """Necessary preparations"""
        self.tokenizer = SoMaJo("de_CMC")

    def _equal_text(self, paragraphs, tokenized_sentences, parallel=1):
        sentences = self.tokenizer.tokenize_text(paragraphs, parallel=parallel)
        sentences = [[t.text for t in s] for s in sentences]
        self.assertEqual(sentences, [ts.split() for ts in tokenized_sentences])

    def _equal_text_file_single_newlines(self, paragraphs, tokenized_sentences, parallel=1):
        pseudofile = io.StringIO("\n".join(paragraphs))
        sentences = self.tokenizer.tokenize_text_file(pseudofile, paragraph_separator="single_newlines", parallel=parallel)
        sentences = [[t.text for t in s] for s in sentences]
        self.assertEqual(sentences, [ts.split() for ts in tokenized_sentences])

    def _equal_text_file_empty_lines(self, paragraphs, tokenized_sentences, parallel=1):
        pseudofile = io.StringIO("\n\n".join(paragraphs))
        sentences = self.tokenizer.tokenize_text_file(pseudofile, paragraph_separator="empty_lines", parallel=parallel)
        sentences = [[t.text for t in s] for s in sentences]
        self.assertEqual(sentences, [ts.split() for ts in tokenized_sentences])

    def _equal_xml(self, xml, tokenized_sentences, strip_tags=False, parallel=1, prune_tags=None):
        eos_tags = "title h1 h2 h3 h4 h5 h6 p br hr div ol ul dl table".split()
        sentences = self.tokenizer.tokenize_xml(xml, eos_tags, strip_tags=strip_tags, parallel=parallel, prune_tags=prune_tags)
        sentences = [[t.text for t in s] for s in sentences]
        self.assertEqual(sentences, [ts.split() for ts in tokenized_sentences])

    def _equal_xml_file(self, xml, tokenized_sentences, strip_tags=False, parallel=1, prune_tags=None):
        eos_tags = "title h1 h2 h3 h4 h5 h6 p br hr div ol ul dl table".split()
        pseudofile = io.StringIO(xml)
        sentences = self.tokenizer.tokenize_xml_file(pseudofile, eos_tags, strip_tags=strip_tags, parallel=parallel, prune_tags=prune_tags)
        sentences = [[t.text for t in s] for s in sentences]
        self.assertEqual(sentences, [ts.split() for ts in tokenized_sentences])


class TestSoMaJoNoSent(TestSoMaJo):
    def setUp(self):
        """Necessary preparations"""
        self.tokenizer = SoMaJo("de_CMC", split_sentences=False)


class TestText(TestSoMaJo):
    def test_text_01(self):
        self._equal_text(["Foo bar. Baz qux", "alpha. Beta gamma"], ["Foo bar .", "Baz qux", "alpha .", "Beta gamma"])

    def test_text_02(self):
        self._equal_text_file_empty_lines(["Foo bar. Baz qux", "alpha. Beta gamma"], ["Foo bar .", "Baz qux", "alpha .", "Beta gamma"])

    def test_text_03(self):
        self._equal_text_file_single_newlines(["Foo bar. Baz qux", "alpha. Beta gamma"], ["Foo bar .", "Baz qux", "alpha .", "Beta gamma"])

    def test_text_04(self):
        self.assertRaises(TypeError, self.tokenizer.tokenize_text, "Foo bar. Baz qux")


class TestTextXMLSent(TestSoMaJo):
    def setUp(self):
        """Necessary preparations"""
        self.tokenizer = SoMaJo("de_CMC", xml_sentences="s")

    def test_text_01(self):
        self._equal_text(["Foo bar. Baz qux", "alpha. Beta gamma"], ["<s> Foo bar . </s>", "<s> Baz qux </s>", "<s> alpha . </s>", "<s> Beta gamma </s>"])

    def test_text_02(self):
        self._equal_text_file_empty_lines(["Foo bar. Baz qux", "alpha. Beta gamma"], ["<s> Foo bar . </s>", "<s> Baz qux </s>", "<s> alpha . </s>", "<s> Beta gamma </s>"])


class TestTextParallel(TestSoMaJo):
    def test_text_01(self):
        self._equal_text(["Foo bar. Baz qux", "alpha. Beta gamma"], ["Foo bar .", "Baz qux", "alpha .", "Beta gamma"], parallel=2)

    def test_text_02(self):
        self._equal_text_file_empty_lines(["Foo bar. Baz qux", "alpha. Beta gamma"], ["Foo bar .", "Baz qux", "alpha .", "Beta gamma"], parallel=2)

    def test_text_03(self):
        self._equal_text_file_single_newlines(["Foo bar. Baz qux", "alpha. Beta gamma"], ["Foo bar .", "Baz qux", "alpha .", "Beta gamma"], parallel=2)


class TestTextNoSent(TestSoMaJoNoSent):
    def test_text_01(self):
        self._equal_text(["Foo bar. Baz qux", "alpha. Beta gamma"], ["Foo bar . Baz qux", "alpha . Beta gamma"])

    def test_text_02(self):
        self._equal_text_file_empty_lines(["Foo bar. Baz qux", "alpha. Beta gamma"], ["Foo bar . Baz qux", "alpha . Beta gamma"])

    def test_text_03(self):
        self._equal_text_file_single_newlines(["Foo bar. Baz qux", "alpha. Beta gamma"], ["Foo bar . Baz qux", "alpha . Beta gamma"])


class TestTextNoSentParallel(TestSoMaJoNoSent):
    def test_text_01(self):
        self._equal_text(["Foo bar. Baz qux", "alpha. Beta gamma"], ["Foo bar . Baz qux", "alpha . Beta gamma"], parallel=2)

    def test_text_02(self):
        self._equal_text_file_empty_lines(["Foo bar. Baz qux", "alpha. Beta gamma"], ["Foo bar . Baz qux", "alpha . Beta gamma"], parallel=2)

    def test_text_03(self):
        self._equal_text_file_single_newlines(["Foo bar. Baz qux", "alpha. Beta gamma"], ["Foo bar . Baz qux", "alpha . Beta gamma"], parallel=2)


class TestXML(TestSoMaJo):
    def test_xml_01(self):
        self._equal_xml("<html>\n  <body>\n    <p>Foo bar. Baz qux</p>\n    <p>alpha. Beta gamma</p>\n  </body>\n</html>", ["<html> <body> <p> Foo bar .", "Baz qux </p>", "<p> alpha .", "Beta gamma </p> </body> </html>"])

    def test_xml_02(self):
        self._equal_xml_file("<html>\n  <body>\n    <p>Foo bar. Baz qux</p>\n    <p>alpha. Beta gamma</p>\n  </body>\n</html>", ["<html> <body> <p> Foo bar .", "Baz qux </p>", "<p> alpha .", "Beta gamma </p> </body> </html>"])


class TestXMLParallel(TestSoMaJo):
    def test_xml_01(self):
        self._equal_xml("<html>\n  <body>\n    <p>Foo bar. Baz qux</p>\n    <p>alpha. Beta gamma</p>\n  </body>\n</html>", ["<html> <body> <p> Foo bar .", "Baz qux </p>", "<p> alpha .", "Beta gamma </p> </body> </html>"], parallel=2)

    def test_xml_02(self):
        self._equal_xml_file("<html>\n  <body>\n    <p>Foo bar. Baz qux</p>\n    <p>alpha. Beta gamma</p>\n  </body>\n</html>", ["<html> <body> <p> Foo bar .", "Baz qux </p>", "<p> alpha .", "Beta gamma </p> </body> </html>"], parallel=2)


class TestXMLNoSent(TestSoMaJoNoSent):
    def test_xml_01(self):
        self._equal_xml("<html>\n  <body>\n    <p>Foo bar. Baz qux</p>\n    <p>alpha. Beta gamma</p>\n  </body>\n</html>", ["<html> <body> <p> Foo bar . Baz qux </p>", "<p> alpha . Beta gamma </p> </body> </html>"])

    def test_xml_02(self):
        self._equal_xml_file("<html>\n  <body>\n    <p>Foo bar. Baz qux</p>\n    <p>alpha. Beta gamma</p>\n  </body>\n</html>", ["<html> <body> <p> Foo bar . Baz qux </p>", "<p> alpha . Beta gamma </p> </body> </html>"])


class TestXMLNoSentParallel(TestSoMaJoNoSent):
    def test_xml_01(self):
        self._equal_xml("<html>\n  <body>\n    <p>Foo bar. Baz qux</p>\n    <p>alpha. Beta gamma</p>\n  </body>\n</html>", ["<html> <body> <p> Foo bar . Baz qux </p>", "<p> alpha . Beta gamma </p> </body> </html>"], parallel=2)

    def test_xml_02(self):
        self._equal_xml_file("<html>\n  <body>\n    <p>Foo bar. Baz qux</p>\n    <p>alpha. Beta gamma</p>\n  </body>\n</html>", ["<html> <body> <p> Foo bar . Baz qux </p>", "<p> alpha . Beta gamma </p> </body> </html>"], parallel=2)


class TestXMLStripTags(TestSoMaJo):
    def test_xml_01(self):
        self._equal_xml("<html>\n  <body>\n    <p>Foo bar. Baz qux</p>\n    <p>alpha. Beta gamma</p>\n  </body>\n</html>", ["Foo bar .", "Baz qux", "alpha .", "Beta gamma"], strip_tags=True)

    def test_xml_02(self):
        self._equal_xml_file("<html>\n  <body>\n    <p>Foo bar. Baz qux</p>\n    <p>alpha. Beta gamma</p>\n  </body>\n</html>", ["Foo bar .", "Baz qux", "alpha .", "Beta gamma"], strip_tags=True)


class TestXMLPruneTags(TestSoMaJo):
    def test_xml_01(self):
        self._equal_xml("<html>\n  <head>\n    Spam\n  </head>\n  <body>\n    <p>Foo bar. Baz qux</p>\n    <p>alpha. Beta gamma</p>\n  </body>\n</html>", ["<html> <body> <p> Foo bar .", "Baz qux </p>", "<p> alpha .", "Beta gamma </p> </body> </html>"], prune_tags=["head"])

    def test_xml_02(self):
        self._equal_xml_file("<html>\n  <head>\n    Spam\n  </head>\n  <body>\n    <p>Foo bar. Baz qux</p>\n    <p>alpha. Beta gamma</p>\n  </body>\n</html>", ["<html> <body> <p> Foo bar .", "Baz qux </p>", "<p> alpha .", "Beta gamma </p> </body> </html>"], prune_tags=["head"])


class TestCharacterOffsets(TestSoMaJo):
    def setUp(self):
        """Necessary preparations"""
        self.tokenizer = SoMaJo("de_CMC", character_offsets=True)

    def _equal_offsets_text_file(self, paragraphs, tokenized_sentences, parallel=1):
        raw = "\n\n".join(paragraphs)
        pseudofile = io.StringIO(raw)
        sentences = self.tokenizer.tokenize_text_file(pseudofile, paragraph_separator="empty_lines", parallel=parallel)
        sentences = list(sentences)
        tokens = [[t.text for t in s] for s in sentences]
        self.assertEqual(tokens, [ts.split() for ts in tokenized_sentences])
        offsets = [[t.character_offset for t in s] for s in sentences]
        extracted = [[raw[s:e] for s, e in sent] for sent in offsets]
        self.assertEqual(tokens, extracted)

    def _equal_offsets_xml(self, xml, tokenized_sentences, strip_tags=False, parallel=1, prune_tags=None):
        eos_tags = "title h1 h2 h3 h4 h5 h6 p br hr div ol ul dl table".split()
        sentences = self.tokenizer.tokenize_xml(xml, eos_tags, strip_tags=strip_tags, parallel=parallel, prune_tags=prune_tags)
        sentences = list(sentences)
        tokens = [[t.text for t in s] for s in sentences]
        self.assertEqual(tokens, [ts.split() for ts in tokenized_sentences])
        offsets = [[t.character_offset for t in s] for s in sentences]
        extracted = [[xml[s:e] for s, e in sent] for sent in offsets]
        self.assertEqual(tokens, extracted)

    def test_text_offsets_01(self):
        self._equal_offsets_text_file(["Foo bar. Baz qux", "alpha. Beta gamma"], ["Foo bar .", "Baz qux", "alpha .", "Beta gamma"])

    def test_xml_offsets_01(self):
        self._equal_offsets_xml("<foo><p>bar</p><p>baz</p></foo>", ["<foo> <p> bar </p>", "<p> baz </p> </foo>"])

    def test_xml_offsets_02(self):
        self._equal_offsets_xml("<foo>\n<p>\nbar\n</p>\n<p>\nbaz\n</p>\n</foo>", ["<foo> <p> bar </p>", "<p> baz </p> </foo>"])



#!/usr/bin/env python3

import operator
import unittest

from somajo.doubly_linked_list import DLL


class TestDLL(unittest.TestCase):
    def test_dll_01(self):
        lst = ["Foo", "", 0, -1, False, True, None]
        dll = DLL(lst)
        self.assertEqual(dll.to_list(), lst)

    def test_dll_02(self):
        lst = ["Foo", "", 0, -1, False, True, None]
        dll = DLL(lst)
        self.assertEqual(DLL(reversed(dll)).to_list(), list(reversed(lst)))

    def test_dll_03(self):
        lst = ["Foo", "", 0, -1, False, True, None]
        dll = DLL(lst)
        self.assertEqual(len(dll), len(lst))

    def test_dll_04(self):
        lst = ["Foo", "", 0, -1, False, True, None]
        dll = DLL(["Foo", "", 0, -1, False, True, None])
        self.assertEqual(str(dll), str(lst))

    def test_dll_05(self):
        dll = DLL([4, 5, 6])
        dll.append_left(3)
        self.assertEqual(dll.to_list(), [3, 4, 5, 6])

    def test_dll_06(self):
        dll = DLL([4, 5, 6])
        dll.append(7)
        self.assertEqual(dll.to_list(), [4, 5, 6, 7])

    def test_dll_07(self):
        dll = DLL([4, 5, 6])
        dll.extend([7, 8, 9])
        self.assertEqual(dll.to_list(), [4, 5, 6, 7, 8, 9])

    def test_dll_08(self):
        dll = DLL([4, 5, 6, 7])
        last = dll.pop()
        self.assertEqual(last, 7)
        self.assertEqual(len(dll), 3)
        self.assertEqual(dll.to_list(), [4, 5, 6])

    def test_dll_09(self):
        dll = DLL([])
        self.assertEqual(len(dll), 0)
        self.assertEqual(dll.to_list(), [])

    def test_dll_10(self):
        dll = DLL([4])
        last = dll.pop()
        self.assertEqual(last, 4)
        self.assertEqual(len(dll), 0)
        self.assertEqual(dll.to_list(), [])

    def test_dll_11(self):
        dll = DLL([4])
        last = dll.pop()
        self.assertEqual(last, 4)
        self.assertRaises(IndexError, dll.pop)

    def test_dll_12(self):
        dll = DLL([])
        dll.append_left(1)
        self.assertEqual(dll.to_list(), [1])

    def test_dll_13(self):
        dll = DLL([1, 2, 3, 4])
        x = dll.next_matching(dll.first, operator.attrgetter("value"), 2)
        self.assertEqual(x.value, 2)
        self.assertEqual([e.value for e in dll.__iter__(start=x)], [2, 3, 4])

    def test_dll_14(self):
        dll = DLL([1, 2, 3, 4])
        x = dll.previous_matching(dll.last, operator.attrgetter("value"), 3)
        self.assertEqual(x.value, 3)
        self.assertEqual([e.value for e in dll.__reversed__(start=x)], [3, 2, 1])

    def test_dll_15(self):
        dll = DLL([1, 2, 3, 4])
        x = dll.next_matching(dll.first, operator.attrgetter("value"), 4, operator.attrgetter("value"), 3)
        self.assertEqual(x.value, 4)

    def test_dll_16(self):
        dll = DLL([1, 2, 3, 4])
        x = dll.next_matching(dll.first, operator.attrgetter("value"), 7)
        self.assertIs(x, None)

    def test_dll_17(self):
        dll = DLL([1, 2, 3])
        x = dll.next_matching(dll.first, operator.attrgetter("value"), 2)
        dll.insert_left(7, x)
        self.assertEqual(dll.to_list(), [1, 7, 2, 3])

    def test_dll_18(self):
        dll = DLL([1, 2, 3])
        x = dll.next_matching(dll.first, operator.attrgetter("value"), 2)
        dll.insert_right(7, x)
        self.assertEqual(dll.to_list(), [1, 2, 7, 3])

    def test_dll_19(self):
        dll = DLL([1, 2, 3])
        self.assertTrue(dll.is_left_of(dll.first, dll.last))

    def test_dll_20(self):
        dll = DLL([1, 2, 3])
        self.assertTrue(dll.is_right_of(dll.last, dll.first))

    def test_dll_21(self):
        dll = DLL([1, 2, 3])
        x = dll.next_matching(dll.first, operator.attrgetter("value"), 2)
        dll.remove(x)
        self.assertEqual(dll.to_list(), [1, 3])

    def test_dll_22(self):
        dll = DLL([1, 2, 3])
        dll.remove(dll.first)
        self.assertEqual(dll.to_list(), [2, 3])

    def test_dll_23(self):
        dll = DLL([1, 2, 3])
        dll.remove(dll.last)
        self.assertEqual(dll.to_list(), [1, 2])

    def test_dll_24(self):
        dll = DLL([1, 2, 3])
        dll.insert_left(0, dll.first)
        self.assertEqual(dll.to_list(), [0, 1, 2, 3])

    def test_dll_25(self):
        dll = DLL([1, 2, 3])
        dll.insert_right(4, dll.last)
        self.assertEqual(dll.to_list(), [1, 2, 3, 4])

    def test_dll_26(self):
        dll = DLL([1, 2, 3])
        self.assertFalse(dll.is_left_of(dll.last, dll.first))

    def test_dll_27(self):
        dll = DLL([1])
        dll.remove(dll.last)
        self.assertEqual(dll.to_list(), [])

    def test_dll_28(self):
        dll = DLL([1])
        dll.remove(dll.first)
        self.assertEqual(dll.to_list(), [])



#!/usr/bin/env python3

import unittest

from somajo import utils


class TestXmlChunkGenerator(unittest.TestCase):
    def _equal(self, raw, chunks, prune_tags=None):
        eos_tags = set(["p"])
        if prune_tags is not None:
            prune_tags = set(prune_tags)
        chunk_info = list(utils.xml_chunk_generator(raw, is_file=False, eos_tags=eos_tags, prune_tags=prune_tags))
        chunk_lists = (ci[0] for ci in chunk_info)
        chunk_lists = [[t.text for t in gc] for gc in chunk_lists]
        self.assertEqual(chunk_lists, chunks)

    def _equal_offsets(self, raw, chunks, prune_tags=None):
        eos_tags = set(["p"])
        if prune_tags is not None:
            prune_tags = set(prune_tags)
        chunk_info = list(utils.xml_chunk_generator(raw, is_file=False, eos_tags=eos_tags, prune_tags=prune_tags, character_offsets=True))
        chunk_lists, raws, positions = zip(*chunk_info)
        offsets = [[t.character_offset for t in cl] for cl in chunk_lists]
        extracted_chunks = [[raw[s:e] for s, e in o] for o in offsets]
        self.assertEqual(extracted_chunks, chunks)

    def test_xml_chunk_generator_01(self):
        self._equal("<x>foo bar</x>", [["<x>", "foo bar", "</x>"]])

    def test_xml_chunk_generator_02(self):
        self._equal("<x><p>foo</p><p>bar</p></x>", [["<x>", "<p>", "foo", "</p>"], ["<p>", "bar", "</p>", "</x>"]])

    def test_xml_chunk_generator_03(self):
        self._equal("<x>\n<p>\nfoo\n</p>\n<p>\nbar\n</p>\n</x>", [["<x>", "\n", "<p>", "\nfoo\n", "</p>"], ["\n", "<p>", "\nbar\n", "</p>", "\n", "</x>"]])

    def test_xml_chunk_generator_04(self):
        self._equal(
            "<x>\n  <p>\n    foo\n  </p>\n  <p>\n    bar\n  </p>\n</x>",
            [["<x>", "\n  ", "<p>", "\n    foo\n  ", "</p>"], ["\n  ", "<p>", "\n    bar\n  ", "</p>", "\n", "</x>"]]
        )

    def test_xml_chunk_generator_05(self):
        self._equal(
            "<x><p>foo</p><i>baz</i><p>bar</p><i>baz</i></x>",
            [["<x>", "<p>", "foo", "</p>"], ["<i>", "baz", "</i>"], ["<p>", "bar", "</p>"], ["<i>", "baz", "</i>", "</x>"]]
        )

    def test_xml_chunk_generator_06(self):
        self._equal(
            "<x><p>foo</p><br/><p>bar</p><br/></x>",
            [["<x>", "<p>", "foo", "</p>"], ["<br>", "</br>", "<p>", "bar", "</p>"], ["<br>", "</br>", "</x>"]]
        )

    def test_xml_chunk_generator_07(self):
        self._equal("<x><del>foo</del><i>bar</i></x>", [["<x>", "<i>", "bar", "</i>", "</x>"]], prune_tags=["del"])

    def test_xml_chunk_generator_08(self):
        self._equal("<x><del>foo</del><p>bar</p></x>", [["<x>", "<p>", "bar", "</p>", "</x>"]], prune_tags=["del"])

    def test_xml_chunk_generator_09(self):
        self._equal("<x>bar\n  <del>foo</del>\nbaz</x>", [["<x>", "bar\n  \nbaz", "</x>"]], prune_tags=["del"])

    def test_xml_chunk_offsets_01(self):
        self._equal_offsets("<foo>T&#x0065;st</foo>", [["<foo>", "T&#x0065;st", "</foo>"]])

    def test_xml_chunk_offsets_02(self):
        self._equal_offsets("<foo>3 &#x003c; 5</foo>", [["<foo>", "3 &#x003c; 5",  "</foo>"]])

    def test_xml_chunk_offsets_03(self):
        self._equal_offsets("<foo>Test&#x00ad;fall</foo>", [["<foo>", "Test&#x00ad;fall", "</foo>"]])

    def test_xml_chunk_offsets_04(self):
        self._equal_offsets("<foo>Test­fall</foo>", [["<foo>", "Test­fall", "</foo>"]])

    def test_xml_chunk_offsets_05(self):
        """Single combining mark"""
        self._equal_offsets("<foo>foo xA&#x0308;x foo</foo>", [["<foo>", "foo xA&#x0308;x foo", "</foo>"]])

    def test_xml_chunk_offsets_06(self):
        """Multiple combining marks"""
        self._equal_offsets("<foo>foo xs&#x0323;&#x0307;x foo</foo>", [["<foo>", "foo xs&#x0323;&#x0307;x foo", "</foo>"]])

    def test_xml_chunk_offsets_07(self):
        """Multiple combining marks"""
        self._equal_offsets("<foo>foo xs&#x0307;&#x0323;x foo</foo>", [["<foo>", "foo xs&#x0307;&#x0323;x foo", "</foo>"]])

    def test_xml_chunk_offsets_08(self):
        """Multiple combining marks"""
        self._equal_offsets("<foo>foo xs&#x1e0b;&#x0323;x foo</foo>", [["<foo>", "foo xs&#x1e0b;&#x0323;x foo", "</foo>"]])

    def test_xml_chunk_offsets_09(self):
        """Multiple combining marks"""
        self._equal_offsets("<foo>foo xq&#x0307;&#x0323;x foo</foo>", [["<foo>", "foo xq&#x0307;&#x0323;x foo", "</foo>"]])

    def test_xml_chunk_offsets_10(self):
        self._equal_offsets("<foo bar='baz'>Foo</foo>", [["<foo bar='baz'>", "Foo", "</foo>"]])

    def test_xml_chunk_offsets_11(self):
        self._equal_offsets("<foo bar='ba\"z'>Foo</foo>", [["<foo bar='ba\"z'>", "Foo", "</foo>"]])

    def test_xml_chunk_offsets_12(self):
        self._equal_offsets("<foo   bar   =   'baz'>   Foo   </foo>", [["<foo   bar   =   'baz'>", "   Foo   ", "</foo>"]])

    def test_xml_chunk_offsets_13(self):
        self._equal_offsets("<foo bar='ba\"z'>Foo \"Bar\" 'Baz'</foo>", [["<foo bar='ba\"z'>", "Foo \"Bar\" 'Baz'", "</foo>"]])

    def test_xml_chunk_offsets_14(self):
        self._equal_offsets('<foo bar="baz"\n  spam="eggs">\n    Foo\n</foo>', [['<foo bar="baz"\n  spam="eggs">', "\n    Foo\n", "</foo>"]])

    def test_xml_chunk_offsets_15(self):
        self._equal_offsets("<foo>Hallo<br/>Tschüß</foo>", [["<foo>", "Hallo", "<br/>", "", "Tschüß", "</foo>"]])

    def test_xml_chunk_offsets_16(self):
        self._equal_offsets("<foo>Hallo<br />Tschüß</foo>", [["<foo>", "Hallo", "<br />", "", "Tschüß", "</foo>"]])

    def test_xml_chunk_offsets_17(self):
        self._equal_offsets("<foo>\u0303foo</foo>", [["<foo>", "\u0303foo", "</foo>"]])

    def test_xml_chunk_offsets_18(self):
        self._equal_offsets("<foo>foo<p>bar</p></foo>", [["<foo>", "foo"], ["<p>", "bar", "</p>", "</foo>"]])

    @unittest.expectedFailure
    def test_xml_chunk_offsets_19(self):
        self._equal_offsets("<foo>bar <del>futsch</del> baz</foo>", [["<foo>", "bar  baz", "</foo>"]], prune_tags=["del"])



#!/usr/bin/env python3

import argparse
import logging
import time

from . import (
    SoMaJo,
    __version__
)

logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.INFO)


def arguments():
    """"""
    parser = argparse.ArgumentParser(description="A tokenizer and sentence splitter for German and English texts. Currently, two tokenization guidelines are implemented: The EmpiriST guidelines for German web and social media texts (de_CMC) and the \"new\" Penn Treebank conventions for English texts (en_PTB).")
    parser.add_argument("-l", "--language", choices=SoMaJo.supported_languages, default=SoMaJo._default_language, help="Choose a language. Currently supported are German EmpiriST-style tokenization (de_CMC) and English Penn-Treebank-style tokenization(en_PTB). (Default: de_CMC)")
    parser.add_argument("-s", "--paragraph_separator", choices=SoMaJo.paragraph_separators, default=SoMaJo._default_parsep, help="How are paragraphs separated in the input text? Will be ignored if option -x/--xml is used. (Default: empty_lines)")
    parser.add_argument("-x", "--xml", action="store_true", help="The input is an XML file. You can specify tags that always constitute a sentence break (e.g. HTML p tags) via the --tag option.")
    parser.add_argument("--tag", action="append", help="Start and end tags of this type constitute sentence breaks, i.e. they do not occur in the middle of a sentence. Can be used multiple times to specify multiple tags, e.g. --tag p --tag br. Implies option -x/--xml. (Default: --tag title --tag h1 --tag h2 --tag h3 --tag h4 --tag h5 --tag h6 --tag p --tag br --tag hr --tag div --tag ol --tag ul --tag dl --tag table)")
    parser.add_argument("--prune", action="append", help="Tags of this type will be removed from the input before tokenization. Can be used multiple times to specify multiple tags, e.g. --tag script --tag style. Implies option -x/--xml. By default, no tags are pruned.")
    parser.add_argument("--strip-tags", action="store_true", help="Suppresses output of XML tags. Implies option -x/--xml.")
    parser.add_argument("-c", "--split_camel_case", action="store_true", help="Split items in written in camelCase (excluding established names and terms).")
    parser.add_argument("--split_sentences", "--split-sentences", action="store_true", help="Also split the input into sentences.")
    parser.add_argument("--sentence_tag", "--sentence-tag", type=str, help="Tag name for sentence boundaries (e.g. --sentence_tag s). If this option is specified, sentences will be delimited by XML tags (e.g. <s>…</s>) instead of empty lines. This option implies --split_sentences")
    parser.add_argument("-t", "--token_classes", action="store_true", help="Output the token classes (number, XML tag, abbreviation, etc.) in addition to the tokens.")
    parser.add_argument("-e", "--extra_info", action="store_true", help='Output additional information for each token: SpaceAfter=No if the token was not followed by a space and OriginalSpelling="…" if the token contained whitespace.')
    parser.add_argument("--character-offsets", action="store_true", help='Output character offsets in the input for each token.')
    parser.add_argument("--parallel", type=int, default=1, metavar="N", help="Run N worker processes (up to the number of CPUs) to speed up tokenization.")
    parser.add_argument("-v", "--version", action="version", version="SoMaJo %s" % __version__, help="Output version information and exit.")
    parser.add_argument("FILE", type=argparse.FileType("r", encoding="utf-8"), help="The input file (UTF-8-encoded) or \"-\" to read from STDIN.")
    args = parser.parse_args()
    return args


def main():
    args = arguments()
    n_tokens = 0
    n_sentences = 0
    t0 = time.perf_counter()
    is_xml = False
    if args.xml or args.strip_tags or (args.tag is not None) or (args.prune is not None):
        is_xml = True
    if args.sentence_tag:
        args.split_sentences = True
    tokenizer = SoMaJo(
        args.language,
        split_camel_case=args.split_camel_case,
        split_sentences=args.split_sentences,
        xml_sentences=args.sentence_tag,
        character_offsets=args.character_offsets
    )
    if is_xml:
        eos_tags = args.tag
        if eos_tags is None:
            eos_tags = "title h1 h2 h3 h4 h5 h6 p br hr div ol ul dl table".split()
        chunks = tokenizer.tokenize_xml_file(args.FILE, eos_tags, strip_tags=args.strip_tags, parallel=args.parallel, prune_tags=args.prune)
    else:
        chunks = tokenizer.tokenize_text_file(args.FILE, args.paragraph_separator, parallel=args.parallel)
    for chunk in chunks:
        n_sentences += 1
        for token in chunk:
            output = token.text
            if not token.markup:
                n_tokens += 1
                if args.token_classes:
                    output += "\t" + token.token_class
                if args.extra_info:
                    output += "\t" + token.extra_info
                if args.character_offsets:
                    output += f"\t{token.character_offset[0]}, {token.character_offset[1]}"
            print(output)
        if args.split_sentences and args.sentence_tag is None:
            print()
    t1 = time.perf_counter()
    if args.split_sentences:
        logging.info("Tokenized %d tokens (%d sentences) in %d seconds (%d tokens/s)" % (n_tokens, n_sentences, t1 - t0, n_tokens / (t1 - t0)))
    else:
        logging.info("Tokenized %d tokens in %d seconds (%d tokens/s)" % (n_tokens, t1 - t0, n_tokens / (t1 - t0)))



#!/usr/bin/env python3

import unicodedata

import regex as re


_ranges = [
    (0x0000, 0x001F),
    (0x007F, 0x009F),
    (0x2000, 0x200A),           # whitespace
    (0x200B, 0x200F),
    (0x202A, 0x202E),
    (0x2066, 0x2069)
]
_single_characters = ["\u00AD", "\u061C", "\u2060", "\uFEFF", "\uFE0F"]
_whitespace = [" ", "\u00A0", "\u1680", "\u2028", "\u2029", "\u202F", "\u205F", "\u3000"]
_skipable_characters = set(_single_characters + _whitespace + [chr(i) for start, end in _ranges for i in range(start, end + 1)])

_xml_entity = re.compile(r"&(?:#\d+|#x[0-9a-f]+|amp|apos|gt|lt|quot);", re.I)


def _align_nfc(nfc, orig):
    """Character alignment from NFC version to original string."""
    alignment = {}
    if nfc == "":
        assert orig == "", "NFC string is empty - expected original string to be also empty; it is '{orig}' instead"
        return alignment
    if nfc == orig:
        return {(i, i + 1): (i, i + 1) for i in range(len(nfc))}
    nfc_i, nfc_j = 0, 0
    orig_i, orig_j = 0, 0
    while nfc_j < len(nfc):
        nfc_j = nfc_i + 1
        while (nfc_j < len(nfc)) and (unicodedata.combining(nfc[nfc_j]) > 0):
            nfc_j += 1
        orig_j = orig_i + 1
        while (orig_j < len(orig)) and (unicodedata.combining(orig[orig_j]) > 0):
            orig_j += 1
        assert nfc[nfc_i:nfc_j] == unicodedata.normalize("NFC", orig[orig_i:orig_j]), f"'{nfc[nfc_i:nfc_j]}' != unicodedata.normalize('NFC', '{orig[orig_i:orig_j]}')"
        alignment[(nfc_i, nfc_j)] = (orig_i, orig_j)
        nfc_i = nfc_j
        orig_i = orig_j
    assert orig_j == len(orig), f"{orig_j} != {len(orig)}; nfc: '{nfc}', orig: '{orig}'"
    return alignment


def _determine_offsets(tokens, raw, position):
    """Determine start and end positions of tokens in the original raw (NFC) input."""
    offsets = []
    raw_i = 0
    raw = re.sub(r"\s", " ", raw)
    for token in tokens:
        if token.markup:
            start, end = token.character_offset
            start -= position
            end -= position
        else:
            text = token.text
            if token.original_spelling is not None:
                text = token.original_spelling
            text = re.sub(r"\s", " ", text)
            if raw[raw_i:].startswith(text):
                start = raw_i
                end = start + len(text)
            elif raw[raw_i:].startswith(" " + text):
                start = raw_i + 1
                end = start + len(text)
            else:
                raw_start = raw_i
                for i, char in enumerate(text):
                    for j in range(raw_start, len(raw)):
                        if raw[j] == char:
                            if i == 0:
                                start = j
                            if i == len(text) - 1:
                                end = j + 1
                            break
                        else:
                            assert raw[j] in _skipable_characters, f"'{raw[j]}' ({hex(ord(raw[j]))}) is not a skipable character; token: '{text}', raw: '{raw[raw_i:]}'"
                    raw_start = j + 1
        offsets.append((start, end))
        raw_i = end
    return offsets


def _resolve_entities(xml):
    """Resolve XML entities and provide an alignment from output string to input string."""
    named = {"&amp;": "&", "&apos;": "'", "&gt;": ">", "&lt;": "<", "&quot;": '"'}
    outstring = ""
    alignment = []
    xml_lower = xml.lower()
    i = 0
    for m in _xml_entity.finditer(xml_lower):
        start, end = m.span()
        if xml_lower[start + 2] == "x":
            char = chr(int(xml[start + 3:end - 1], base=16))
        elif xml_lower[start + 1] == "#":
            char = chr(int(xml[start + 2:end - 1]))
        else:
            char = named[xml_lower[start:end]]
        outstring += xml[i:start] + char
        for j in range(i, start):
            alignment.append((j, j + 1))
        alignment.append((start, end))
        i = end
    outstring += xml[i:len(xml)]
    for j in range(i, len(xml)):
        alignment.append((j, j + 1))
    return outstring, alignment


def token_offsets(token_list, raw, position, xml_input, tokens):
    """Determine character offsets for tokens."""
    if xml_input:
        chunk_offsets = [(t.character_offset[0] - position, t.character_offset[1] - position) for t in token_list]
        raw, align_to_entities = _resolve_entities(raw)
        align_from_entities = {i: char_i for char_i, (start, end) in enumerate(align_to_entities) for i in range(start, end)}
        chunks = [raw[align_from_entities[start]:align_from_entities[end - 1] + 1] for start, end in chunk_offsets]
        chunks_nfc = [unicodedata.normalize("NFC", c) for c in chunks]
        alignments = [_align_nfc(chunk_nfc, chunk) for chunk, chunk_nfc in zip(chunks, chunks_nfc)]
        align_to_raw = alignments[0]
        for i in range(1, len(alignments)):
            o1 = sum(len(c) for c in chunks_nfc[:i])
            o2 = sum(len(c) for c in chunks[:i])
            align_to_raw.update({(k[0] + o1, k[1] + o1): (v[0] + o2, v[1] + o2) for k, v in alignments[i].items()})
        raw_nfc = "".join(chunks_nfc)
    else:
        raw_nfc = unicodedata.normalize("NFC", raw)
        align_to_raw = _align_nfc(raw_nfc, raw)
    align_from_raw = {i: k for k, v in align_to_raw.items() for i in range(v[0], v[1])}
    align_to_starts = {i: v[0] for k, v in align_to_raw.items() for i in range(k[0], k[1])}
    align_to_ends = {i: v[1] for k, v in align_to_raw.items() for i in range(k[0], k[1])}
    # adjust character offsets for markup tokens
    if xml_input:
        for i in range(len(tokens)):
            if tokens[i].markup:
                s, e = tokens[i].character_offset
                tokens[i].character_offset = (
                    align_from_raw[align_from_entities[s - position]][0] + position,
                    align_from_raw[align_from_entities[e - position - 1]][1] + position
                )
    offsets = _determine_offsets(tokens, raw_nfc, position)
    assert len(tokens) == len(offsets), f"Not as many tokens as offsets: {len(tokens)} != {len(offsets)}"
    offsets = [(align_to_starts[s], align_to_ends[e - 1]) for s, e in offsets]
    if xml_input:
        offsets = [(align_to_entities[s][0], align_to_entities[e - 1][1]) for s, e in offsets]
    offsets = [(s + position, e + position) for s, e in offsets]
    return offsets


def xml_chunk_offset(token, raw):
    """Determine character offset for an XML chunk created by `utils._xml_chunk_generator`."""
    raw, align_to_raw = _resolve_entities(raw)
    raw = re.sub(r"\s", " ", raw)
    text = token.text
    text = re.sub(r"\s", " ", text)
    if token.markup:
        text, align_to_text = _resolve_entities(text)
        text = text.replace("'", '"')
        if raw.startswith(text):
            start = 0
            end = len(text)
        else:
            pattern = "(" + re.escape(text) + ")"
            pattern = pattern.replace(r"\ ", r"\s+")
            pattern = pattern.replace("=", r"\s*=\s*")
            if not text.startswith("</"):
                pattern = pattern[:-2] + r"\s*/?\s*" + pattern[-2:]
            local_raw = raw.replace("'", '"')
            m = re.match(pattern, local_raw)
            if text.startswith("</") and not m:
                start, end = 0, 0
            else:
                assert m, f"'{text}' not found in '{local_raw}'"
                start, end = m.span(1)
    else:
        assert raw.startswith(text), f"'{raw}' does not start with '{text}'"
        start = 0
        end = len(text)
    if start == end:
        return (align_to_raw[start][0], align_to_raw[start][0])
    else:
        return (align_to_raw[start][0], align_to_raw[end - 1][1])



#!/usr/bin/env python3

import functools
import itertools
import multiprocessing

from . import (
    alignment,
    doubly_linked_list,
    utils
)
from .sentence_splitter import SentenceSplitter
from .token import Token
from .tokenizer import Tokenizer


class SoMaJo:
    """Tokenization and sentence splitting.

    Parameters
    ----------
    language : {'de_CMC', 'en_PTB'}
        Language-specific tokenization rules.
    split_camel_case : bool, (default=False)
        Split words written in camelCase (excluding established names
        and terms).
    split_sentences : bool, (default=True)
        Perform sentence splitting in addition to tokenization.
    xml_sentences : str, (default=None)
        Delimit sentences by XML tags of this name
        (``xml_sentences='s'`` → <s>…</s>). When used with XML input,
        this might lead to minor changes to the original tags to
        guarantee well-formed output (tags might need to be closed and
        re-opened at sentence boundaries).
    character_offsets : bool, (default=False)
        Compute the character offsets in the input for each token.
        This allows for stand-off tokenization.

    """

    supported_languages = {"de_CMC", "en_PTB"}
    _default_language = "de_CMC"
    paragraph_separators = {"empty_lines", "single_newlines"}
    _default_parsep = "empty_lines"

    def __init__(self, language, *, split_camel_case=False, split_sentences=True, xml_sentences=None, character_offsets=False):
        assert language in self.supported_languages
        self.language = language
        self.split_camel_case = split_camel_case
        self.split_sentences = split_sentences
        self.xml_sentences = xml_sentences
        self.character_offsets = character_offsets
        self._tokenizer = Tokenizer(split_camel_case=self.split_camel_case, language=self.language)
        if self.split_sentences:
            self._sentence_splitter = SentenceSplitter(language=self.language)

    def _tokenize(self, token_info, xml_input):
        """Tokenize and sentence split a single token_dll."""
        token_list, raw, position = token_info
        token_dll = doubly_linked_list.DLL(token_list)
        tokens = self._tokenizer._tokenize(token_dll)
        if self.character_offsets:
            offsets = alignment.token_offsets(token_list, raw, position, xml_input, tokens)
            for i in range(len(tokens)):
                tokens[i].character_offset = offsets[i]
        if self.split_sentences:
            tokens = self._sentence_splitter._split_sentences(tokens)
        return tokens

    def _parallel_tokenize(self, token_info, *, parallel=1, strip_tags=False, xml_input=False):
        """Tokenize and sentence split an iterable of token_dlls; optional
        parallelization.

        """
        def partok():
            with multiprocessing.Pool(min(parallel, multiprocessing.cpu_count())) as pool:
                tokens = pool.imap(
                    functools.partial(self._tokenize, xml_input=xml_input),
                    token_info,
                    250
                )
                for par in tokens:
                    yield par

        if parallel > 1:
            tokens = partok()
        else:
            tokens = map(
                functools.partial(self._tokenize, xml_input=xml_input),
                token_info
            )
        if self.split_sentences:
            tokens = itertools.chain.from_iterable(tokens)
            tokens = self._sentence_splitter._merge_empty_sentences(tokens)
        if strip_tags:
            tokens = ([t for t in par if not t.markup] for par in tokens)
        if self.split_sentences and (self.xml_sentences is not None):
            tokens = self._sentence_splitter._add_xml_tags(tokens, s_tag=self.xml_sentences)
        return tokens

    def _tokenize_text(self, token_info, parallel):
        tokens = self._parallel_tokenize(token_info, parallel=parallel)
        if self.xml_sentences:
            tokens = map(utils.escape_xml_tokens, tokens)
        return tokens

    def _tokenize_xml(self, xml_data, is_file, eos_tags, strip_tags, parallel, prune_tags):
        if eos_tags is not None:
            eos_tags = set(eos_tags)
        if prune_tags is not None:
            prune_tags = set(prune_tags)
            assert not self.character_offsets, "Cannot use `prune_tags` when SoMaJo is initialized with `character_offsets=True`."
        token_info = utils.xml_chunk_generator(
            xml_data,
            is_file,
            eos_tags=eos_tags,
            prune_tags=prune_tags,
            character_offsets=self.character_offsets
        )
        tokens = self._parallel_tokenize(token_info, parallel=parallel, strip_tags=strip_tags, xml_input=True)
        if not (strip_tags and self.xml_sentences is None):
            tokens = map(utils.escape_xml_tokens, tokens)
        return tokens

    def tokenize_text_file(self, text_file, paragraph_separator, *, parallel=1):
        """Split the contents of a text file into sequences of tokens.

        Parameters
        ----------
        text_file : str or file-like object
            Either a filename or a file-like object containing text.
        paragraph_separator : {'single_newlines', 'empty_lines'}
            How are paragraphs separated in the input? Is there one
            paragraph per line ('single_newlines') or do paragraphs
            span several lines and are separated by 'empty_lines'?
        parallel : int, (default=1)
            Number of processes to use.

        Yields
        -------
        list
            The ``Token`` objects in a single sentence or paragraph
            (depending on the value of ``split_sentences``).

        Examples
        --------

        Tokenization and sentence splitting; input file with
        paragraphs separated by empty lines; print one token per line
        with token classes and extra information; print an empty line
        after each sentence:

        >>> with open("example_empty_lines.txt") as f:
        ...     print(f.read())
        ... 
        Heyi:)
        ​
        Was machst du morgen Abend?! Lust auf Film?;-)
        >>> sentences = tokenizer.tokenize_text_file("example_empty_lines.txt", paragraph_separator="single_newlines")
        >>> for sentence in sentences:
        ...     for token in sentence:
        ...         print("{token.text}\t{token.token_class}\t{token.extra_info}")
        ...     print()
        ... 
        Heyi	regular	SpaceAfter=No
        :)	emoticon	
        ​
        Was	regular	
        machst	regular	
        du	regular	
        morgen	regular	
        Abend	regular	SpaceAfter=No
        ?!	symbol	
        ​
        Lust	regular	
        auf	regular	
        Film	regular	SpaceAfter=No
        ?	symbol	SpaceAfter=No
        ;-)	emoticon	
        ​

        Tokenization and sentence splitting; input file with
        paragraphs separated by single newlines; print one sentence
        per line:

        >>> with open("example_single_newlines.txt", encoding="utf-8") as f:
        ...     print(f.read())
        ... 
        Heyi:)
        Was machst du morgen Abend?! Lust auf Film?;-)
        >>> tokenizer = SoMaJo("de_CMC")
        >>> with open("example_empty_lines.txt", encoding="utf-8") as f:
        ...     sentences = tokenizer.tokenize_text_file(f, paragraph_separator="empty_lines")
        ...     for sentence in sentences:
        ...         print(" ".join([token.text for token in sentence]))
        ... 
        Heyi :)
        Was machst du morgen Abend ?!
        Lust auf Film ? ;-)

        """
        assert paragraph_separator in self.paragraph_separators
        token_info = utils.get_paragraphs_list(text_file, paragraph_separator)
        return self._tokenize_text(token_info, parallel)

    def tokenize_xml_file(self, xml_file, eos_tags, *, strip_tags=False, parallel=1, prune_tags=None):
        """Split the contents of an xml file into sequences of tokens.

        Parameters
        ----------
        xml_file : str or file-like object
            A file containing XML data. Either a filename or a
            file-like object.
        eos_tags : iterable
            XML tags that constitute sentence breaks, i.e. tags that
            do not occur in the middle of a sentence. For HTML input,
            you might use the following list of tags: ``['title',
            'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'p', 'br', 'hr',
            'div', 'ol', 'ul', 'dl', 'table']``
        strip_tags : bool, (default=False)
            Remove all XML tags from the output.
        parallel : int, (default=1)
            Number of processes to use.
        prune_tags : iterable
            These XML tags and their contents will be removed from the
            input before tokenization. For HTML input, you might use
            ``['script', 'style']`` or, depending on your use case,
            ``['head']``.

        Yields
        -------
        list
            The ``Token`` objects in a single sentence or stretch of
            XML delimited by ``eos_tags`` (depending on the value of
            ``split_sentences``).

        Examples
        --------

        Tokenization and sentence splitting; print one token per line
        and an empty line after each sentence:

        >>> with open("example.xml") as f:
        ...     print(f.read())
        ... 
        <html>
          <body>
            <p>Heyi:)</p>
            <p>Was machst du morgen Abend?! Lust auf Film?;-)</p>
          </body>
        </html>
        >>> eos_tags = "title h1 h2 h3 h4 h5 h6 p br hr div ol ul dl table".split()
        >>> tokenizer = SoMaJo("de_CMC")
        >>> sentences = tokenizer.tokenize_xml_file("example.xml", eos_tags)
        >>> for sentence in sentences:
        ...     for token in sentence:
        ...         print(token)
        ...     print()
        ... 
        <html>
        <body>
        <p>
        Heyi
        :)
        </p>
        ​
        <p>
        Was
        machst
        du
        morgen
        Abend
        ?!
        ​
        Lust
        auf
        Film
        ?
        ;-)
        </p>
        </body>
        </html>
        ​

        Tokenization and sentence splitting; strip XML tags from the
        output and print one sentence per line:

        >>> with open("example.xml") as f:
        ...     sentences = tokenizer.tokenize_xml_file(f, eos_tags, strip_tags=True)
        ...     for sentence in sentences:
        ...         print(" ".join(token.text for token in sentence))
        ... 
        Heyi :)
        Was machst du morgen Abend ?!
        Lust auf Film ? ;-)

        Only tokenization; print one token per line

        >>> tokenizer = SoMaJo("de_CMC", split_sentences=False)
        >>> chunks = tokenizer.tokenize_xml_file("example.xml", eos_tags)
        >>> for chunk in chunks:
        ...     for token in chunk:
        ...         print(token.text)
        ... 
        <html>
        <body>
        <p>
        Heyi
        :)
        </p>
        <p>
        Was
        machst
        du
        morgen
        Abend
        ?!
        Lust
        auf
        Film
        ?
        ;-)
        </p>
        </body>
        </html>

        """
        return self._tokenize_xml(
            xml_file,
            is_file=True,
            eos_tags=eos_tags,
            strip_tags=strip_tags,
            parallel=parallel,
            prune_tags=prune_tags
        )

    def tokenize_text(self, paragraphs, *, parallel=1):
        """Split paragraphs of text into sequences of tokens.

        Parameters
        ----------
        paragraphs : iterable
            An iterable of single paragraphs of text.
        parallel : int, (default=1)
            Number of processes to use.

        Yields
        ------
        list
            The ``Token`` objects in a single sentence or paragraph
            (depending on the value of ``split_sentences``).

        Examples
        --------

        Tokenization and sentence splitting; print one sentence per
        line:

        >>> paragraphs = ["Heyi:)", "Was machst du morgen Abend?! Lust auf Film?;-)"]
        >>> tokenizer = SoMaJo("de_CMC")
        >>> sentences = tokenizer.tokenize_text(paragraphs)
        >>> for sentence in sentences:
        ...     print(" ".join([token.text for token in sentence]))
        ... 
        Heyi :)
        Was machst du morgen Abend ?!
        Lust auf Film ? ;-)

        Only tokenization; print one paragraph per line:

        >>> tokenizer = SoMaJo("de_CMC", split_sentences=False)
        >>> tokenized_paragraphs = tokenizer.tokenize_text(paragraphs)
        >>> for paragraph in tokenized_paragraphs:
        ...     print(" ".join([token.text for token in paragraph]))
        ... 
        Heyi :)
        Was machst du morgen Abend ?! Lust auf Film ? ;-)

        Tokenization and sentence splitting; print one token per line
        with token classes and extra information; print an empty line
        after each sentence:

        >>> sentences = tokenizer.tokenize_text(paragraphs)
        >>> for sentence in sentences:
        ...     for token in sentence:
        ...         print("{token.text}\t{token.token_class}\t{token.extra_info}")
        ...     print()
        ... 
        Heyi	regular	SpaceAfter=No
        :)	emoticon	
        ​
        Was	regular	
        machst	regular	
        du	regular	
        morgen	regular	
        Abend	regular	SpaceAfter=No
        ?!	symbol	
        ​
        Lust	regular	
        auf	regular	
        Film	regular	SpaceAfter=No
        ?	symbol	SpaceAfter=No
        ;-)	emoticon
        ​

        Tokenization and sentence splitting; print one token per line
        and delimit sentences with XML tags:

        >>> tokenizer = SoMaJo("de_CMC", xml_sentences="s")
        >>> sentences = tokenizer.tokenize_text(paragraphs)
        >>> for sentence in sentences:
        ...     for token in sentence:
        ...         print(token.text)
        ... 
        <s>
        Heyi
        :)
        </s>
        <s>
        Was
        machst
        du
        morgen
        Abend
        ?!
        </s>
        <s>
        Lust
        auf
        Film
        ?
        ;-)
        </s>

        """
        if isinstance(paragraphs, str):
            raise TypeError("``paragraphs`` must be an iterable of strings, not a string!")
        token_info = (([Token(p, first_in_sentence=True, last_in_sentence=True)], p, 0) for p in paragraphs)
        return self._tokenize_text(token_info, parallel)

    def tokenize_xml(self, xml_data, eos_tags, *, strip_tags=False, parallel=1, prune_tags=None):
        """Split a string of XML data into sequences of tokens.

        Parameters
        ----------
        xml_data : str
            A string containing XML data.
        eos_tags : iterable
            XML tags that constitute sentence breaks, i.e. tags that
            do not occur in the middle of a sentence. For HTML input,
            you might use the following list of tags: ``['title',
            'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'p', 'br', 'hr',
            'div', 'ol', 'ul', 'dl', 'table']``
        strip_tags : bool, (default=False)
            Remove the XML tags from the output.
        parallel : int, (default=1)
            Number of processes to use.
        prune_tags : iterable
            These XML tags and their contents will be removed from the
            input before tokenization. For HTML input, you might use
            ``['script', 'style']`` or, depending on your use case,
            ``['head']``.

        Yields
        ------
        list
            The ``Token`` objects in a single sentence or stretch of
            XML delimited by ``eos_tags`` (depending on the value of
            ``split_sentences``).

        Examples
        --------

        Tokenization and sentence splitting; print one token per line
        and an empty line after each sentence:

        >>> xml = "<html><body><p>Heyi:)</p><p>Was machst du morgen Abend?! Lust auf Film?;-)</p></body></html>"
        >>> eos_tags = "title h1 h2 h3 h4 h5 h6 p br hr div ol ul dl table".split()
        >>> tokenizer = SoMaJo("de_CMC")
        >>> sentences = tokenizer.tokenize_xml(xml, eos_tags)
        >>> for sentence in sentences:
        ...     for token in sentence:
        ...         print(token.text)
        ...     print()
        ... 
        <html>
        <body>
        <p>
        Heyi
        :)
        </p>
        ​
        <p>
        Was
        machst
        du
        morgen
        Abend
        ?!
        ​
        Lust
        auf
        Film
        ?
        ;-)
        </p>
        </body>
        </html>
        ​

        Tokenization and sentence splitting; strip XML tags from the
        output and print one sentence per line

        >>> sentences = tokenizer.tokenize_xml(xml, eos_tags, strip_tags=True)
        >>> for sentence in sentences:
        ...     print(" ".join([token.text for token in sentence]))
        ... 
        Heyi :)
        Was machst du morgen Abend ?!
        Lust auf Film ? ;-)

        Only tokenization; print one chunk of XML (delimited by
        ``eos_tags``) per line:

        >>> tokenizer = SoMaJo("de_CMC", split_sentences=False)
        >>> chunks = tokenizer.tokenize_xml(xml, eos_tags)
        >>> for chunk in chunks:
        ...     print(" ".join([token.text for token in chunk]))
        ... 
        <html> <body> <p> Heyi :) </p>
        <p> Was machst du morgen Abend ?! Lust auf Film ? ;-) </p> </body> </html>

        Tokenization and sentence splitting; print one token per line
        and delimit sentences with XML tags:

        >>> xml = "<html><body><p>Heyi:)</p><p>Was machst du morgen Abend?! Lust auf Film?;-)</p></body></html>"
        >>> eos_tags = "title h1 h2 h3 h4 h5 h6 p br hr div ol ul dl table".split()
        >>> tokenizer = SoMaJo("de_CMC", xml_sentences="s")
        >>> sentences = tokenizer.tokenize_xml(xml, eos_tags)
        >>> for sentence in sentences:
        ...     for token in sentence:
        ...         print(token.text)
        ...     print()
        ...
        <html>
        <body>
        <p>
        <s>
        Heyi
        :)
        </s>
        </p>
        <p>
        <s>
        Was
        machst
        du
        morgen
        Abend
        ?!
        </s>
        <s>
        Lust
        auf
        Film
        ?
        ;-)
        </s>
        </p>
        </body>
        </html>

        """
        return self._tokenize_xml(
            xml_data,
            is_file=False,
            eos_tags=eos_tags,
            strip_tags=strip_tags,
            parallel=parallel,
            prune_tags=prune_tags
        )
